from pytest import raises

from desmod.queue import PriorityItem, PriorityQueue, Queue


def test_mq(env):
    queue = Queue(env, capacity=2)

    def producer(msg, wait):
        yield env.timeout(wait)
        yield queue.put(msg)

    def consumer(expected_msg, wait):
        yield env.timeout(wait)
        msg = yield queue.get()
        assert msg == expected_msg

    env.process(producer('1st', 0))
    env.process(producer('2nd', 1))
    env.process(consumer('1st', 0))
    env.process(consumer('2nd', 1))
    env.run()


def test_queue_peek(env):
    queue = Queue(env)
    assert queue.is_empty
    with raises(IndexError):
        queue.peek()

    queue2 = Queue(env, items=[9, 8, 7])
    assert not queue2.is_empty
    assert queue2.peek() == 9


def test_queue_overflow(env):
    def proc(env, queue):
        yield queue.put(1)
        yield env.timeout(1)
        yield queue.put(1)
        yield env.timeout(1)
        with raises(OverflowError):
            yield queue.put(1)

    queue = Queue(env, capacity=2, hard_cap=True)
    env.process(proc(env, queue))
    env.run()


def test_mq_when_full(env):
    queue = Queue(env, capacity=2)
    result = []

    def producer(env):
        yield env.timeout(1)
        for i in range(5):
            yield queue.put(i)
            yield env.timeout(1)

    def consumer(env):
        yield env.timeout(5)
        for i in range(3):
            msg = yield queue.get()
            assert msg == i

    def full_waiter(env):
        yield queue.when_full()
        result.append('full')

    def any_waiter(env):
        yield queue.when_any()
        assert env.now == 1
        result.append('any')

    env.process(producer(env))
    env.process(consumer(env))
    env.process(full_waiter(env))
    env.process(any_waiter(env))
    env.process(any_waiter(env))
    env.run()
    assert queue.items
    assert queue.is_full
    assert 'full' in result
    assert result.count('any') == 2


def test_priority_mq(env):
    queue = PriorityQueue(env)

    def producer(env):
        for priority in reversed(range(5)):
            item = set([priority])  # unhashable
            yield queue.put(PriorityItem(priority, item))
            yield env.timeout(1)

    def consumer(env):
        yield env.timeout(5)
        for i in range(5):
            msg = yield queue.get()
            assert msg.item == set([i])
            yield env.timeout(1)

    env.process(producer(env))
    env.process(consumer(env))
    env.run()


def test_queue_repr(env):
    queue = Queue(env, name='hi', items=[3, 2, 1])
    assert str(queue) == "Queue(name='hi' size=3 capacity=inf)"

    pri_queue = PriorityQueue(env, capacity=3)
    assert str(pri_queue) == 'PriorityQueue(name=None size=0 capacity=3)'


def test_when_not_full(env):
    queue = Queue(env, capacity=2, items=[0, 1])

    def consumer(env):
        for i in range(2):
            yield env.timeout(3)
            msg = yield queue.get()
            assert msg == i

    def not_full_waiter(env):
        yield queue.when_not_full()
        assert env.now == 3
        yield queue.when_not_full()
        assert env.now == 3

    env.process(consumer(env))
    env.process(not_full_waiter(env))
    env.run()


def test_when_empty(env):
    def proc(env, queue):
        yield queue.when_empty()
        assert env.now == 0

        yield queue.put('a')
        yield queue.put('b')

        with queue.when_empty() as when_empty_ev:
            assert not when_empty_ev.triggered
            yield env.timeout(1)
            item = yield queue.get()
            assert item == 'a'
            assert not when_empty_ev.triggered

        with queue.when_empty() as when_empty_ev:
            assert not when_empty_ev.triggered
            yield env.timeout(1)
            with queue.get() as get_ev:
                item = yield get_ev
                assert item == 'b'
            assert when_empty_ev.triggered
            yield when_empty_ev

    env.process(proc(env, Queue(env)))
    env.run()


def test_when_at_most(env):
    def proc(env, queue):
        for item in 'abc':
            with queue.put(item) as put_ev:
                yield put_ev

        at_most = {}
        at_most[0] = queue.when_at_most(0)
        at_most[3] = queue.when_at_most(3)
        at_most[1] = queue.when_at_most(1)
        at_most[2] = queue.when_at_most(2)
        assert not at_most[0].triggered
        assert not at_most[1].triggered
        assert not at_most[2].triggered
        assert at_most[3].triggered

        item = yield queue.get()
        assert item == 'a'
        assert not at_most[0].triggered
        assert not at_most[1].triggered
        assert at_most[2].triggered

        item = yield queue.get()
        assert item == 'b'
        assert not at_most[0].triggered
        assert at_most[1].triggered

        item = yield queue.get()
        assert item == 'c'
        assert at_most[0].triggered

    env.process(proc(env, Queue(env)))
    env.run()


def test_when_at_least(env):
    def proc(env, queue):
        at_least = {}
        at_least[3] = queue.when_at_least(3)
        at_least[0] = queue.when_at_least(0)
        at_least[2] = queue.when_at_least(2)
        at_least[1] = queue.when_at_least(1)
        assert at_least[0].triggered
        assert not at_least[1].triggered
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield queue.put('a')
        assert at_least[1].triggered
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield queue.get()
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield queue.put('b')
        assert not at_least[2].triggered
        assert not at_least[3].triggered

        yield queue.put('c')
        assert at_least[2].triggered
        assert not at_least[3].triggered

        yield queue.put('d')
        assert at_least[3].triggered

    env.process(proc(env, Queue(env)))
    env.run()


def test_queue_cancel(env):
    queue = Queue(env, capacity=2)

    def producer(env):
        for i in range(5):
            yield env.timeout(5)
            yield queue.put(i)

    def consumer(env):
        for i in range(3):
            yield env.timeout(10)
            msg = yield queue.get()
            assert msg == i

    def canceller(env):
        any_ev = queue.when_any()
        get_ev = queue.get()
        full_ev = queue.when_full()

        yield env.timeout(1)

        assert not get_ev.triggered
        assert not any_ev.triggered
        assert not full_ev.triggered
        get_ev.cancel()
        any_ev.cancel()
        full_ev.cancel()

        assert not queue.is_full
        with queue.when_full() as when_full:
            yield when_full

        with queue.put(1) as put_ev:
            not_full_ev = queue.when_not_full()

            yield env.timeout(1)

            assert not put_ev.triggered
            assert not not_full_ev.triggered
        not_full_ev.cancel()

        yield env.timeout(100)

        assert not get_ev.triggered
        assert not any_ev.triggered
        assert not put_ev.triggered
        assert not put_ev.triggered
        assert not not_full_ev.triggered

    env.process(producer(env))
    env.process(consumer(env))
    env.process(canceller(env))
    env.run()



"""Full-featured, high-level modeling using `SimPy`__.

__ https://simpy.readthedocs.io/en/latest/contents.html

The `desmod` package provides a variety of tools for composing, configuring,
running, monitoring, and analyzing discrete event simulation (DES) models. It
builds on top of the :mod:`simpy` simulation kernel, providing features useful
for building large-scale models which are out-of-scope for :mod:`simpy` itself.

An understanding of SimPy is required to use desmod effectively.

Components
==========

The primary building-block for `desmod` models is the
:class:`~desmod.component.Component` class. Components provide a means for
partitioning the system to be modeled into manageable pieces. Components can
play a structural role by parenting other components; or play a behavioral role
by having processes and connections to other components; or sometimes play both
roles at once.

The :func:`desmod.dot.component_to_dot()` function may be used to create a
`DOT`__ language representation of the component hierarchy and/or the component
connection graph. The resulting DOT representation may be rendered to a variety
of graphical formats using `GraphViz`__ tools.

__ http://graphviz.org/content/dot-language
__ http://graphviz.org/

Configuration
=============

It is common for models to have configurable paramaters. Desmod provides an
opinionated mechanism for simulation configuration. A single, comprehensive
configuration dictionary captures all configuration for the simulation. The
configuration dictionary is propogated to all Components via the
:class:`~desmod.simulation.SimEnvironment`.

The various components (or component hierarchies) may maintain separate
configuration namespaces within the configuration dictionary by use of keys
conforming to the dot-separated naming convention. For example,
"mymodel.compA.cfgitem".

The :mod:`desmod.config` module provides various functionality useful for
managing configuration dictionaries.

Simulation
==========

Desmod takes care of the details of running simulations to allow focus on the
act of modeling.

Running a simulation is accomplished with either
:func:`~desmod.simulation.simulate()` or
:func:`~desmod.simulation.simulate_factors()`, depending whether running a
single simulation or a multi-factor set of simulations. In either case, the key
ingredients are the configuration dict and the model's top-level
:class:`~desmod.component.Component`. The :func:`~desmod.simulation.simulate()`
function takes responsibility for taking the simulation through its various
phases:

 - *Initialization*: where the components' `__init__()` methods are called.
 - *Elaboration*: where inter-component connections are made and components'
   processes are started.
 - *Simulation*: where discrete event simulation occurs.
 - *Post-simulation*: where simulation results are gathered.

# TODO: simulation results

Monitoring
==========

# TODO: tracers, probes, logging, etc.

"""

__all__ = ()



"""Pool class for modeling a container of resources.

A pool models a container of homogeneous resources, similar to
:class:`simpy.resources.Container`, but with additional events when the
container is empty or full. Resources are :func:`Pool.put` or :func:`Pool.get`
to/from the pool in specified amounts. The pool's resources may be modeled as
either discrete or continuous depending on whether the put/get amounts are
`int` or `float`.
"""

from heapq import heapify, heappop, heappush
from sys import float_info
from types import TracebackType
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Type, Union

from simpy.core import BoundClass, Environment
from simpy.events import Event

EventCallback = Callable[[Event], None]
PoolAmount = Union[int, float]


class PoolPutEvent(Event):
    callbacks: List[EventCallback]

    def __init__(self, pool: 'Pool', amount: PoolAmount = 1) -> None:
        if not (0 < amount <= pool.capacity):
            raise ValueError('amount must be in (0, capacity]')
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        self.callbacks.extend([pool._trigger_when_at_least, pool._trigger_get])
        pool._put_waiters.append(self)
        pool._trigger_put()

    def __enter__(self) -> 'PoolPutEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._put_waiters.remove(self)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PoolGetEvent(Event):
    callbacks: List[EventCallback]

    def __init__(self, pool: 'Pool', amount: PoolAmount = 1) -> None:
        if not (0 < amount <= pool.capacity):
            raise ValueError('amount must be in (0, capacity]')
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        self.callbacks.extend([pool._trigger_when_at_most, pool._trigger_put])
        pool._get_waiters.append(self)
        pool._trigger_get()

    def __enter__(self) -> 'PoolGetEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._get_waiters.remove(self)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PoolWhenAtMostEvent(Event):
    def __init__(self, pool: 'Pool', amount: PoolAmount) -> None:
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        heappush(pool._at_most_waiters, self)
        pool._trigger_when_at_most()

    def __lt__(self, other: 'PoolWhenAtMostEvent') -> bool:
        return self.amount > other.amount

    def __enter__(self) -> 'PoolWhenAtMostEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._at_most_waiters.remove(self)
            heapify(self.pool._at_most_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PoolWhenAtLeastEvent(Event):
    def __init__(self, pool: 'Pool', amount: PoolAmount) -> None:
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        heappush(pool._at_least_waiters, self)
        pool._trigger_when_at_least()

    def __lt__(self, other: 'PoolWhenAtLeastEvent') -> bool:
        return self.amount < other.amount

    def __enter__(self) -> 'PoolWhenAtLeastEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._at_least_waiters.remove(self)
            heapify(self.pool._at_least_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PoolWhenAnyEvent(PoolWhenAtLeastEvent):
    def __init__(self, pool: 'Pool', epsilon: float = float_info.min):
        super().__init__(pool, amount=epsilon)


class PoolWhenFullEvent(PoolWhenAtLeastEvent):
    def __init__(self, pool: 'Pool'):
        super().__init__(pool, amount=pool.capacity)


class PoolWhenNotFullEvent(PoolWhenAtMostEvent):
    def __init__(self, pool: 'Pool', epsilon: Optional[float] = None):
        if epsilon is None and isinstance(pool.capacity, int):
            epsilon = 0.5
        assert epsilon is not None, "when_not_any(epsilon) is required for float Pool."
        super().__init__(pool, amount=pool.capacity - epsilon)


class PoolWhenEmptyEvent(PoolWhenAtMostEvent):
    def __init__(self, pool: 'Pool'):
        super().__init__(pool, amount=0)


class Pool:
    """Simulation pool of discrete or continuous resources.

    `Pool` is similar to :class:`simpy.resources.Container`.
    It provides a simulation-aware container for managing a shared pool of
    resources. The resources can be either discrete objects (like apples) or
    continuous (like water).

    Resources are added and removed using :meth:`put()` and :meth:`get()`.

    :param env: Simulation environment.
    :param capacity: Capacity of the pool; infinite by default.
    :param hard_cap:
        If specified, the pool overflows when the `capacity` is reached.
    :param init_level: Initial level of the pool.
    :param name: Optional name to associate with the queue.

    """

    def __init__(
        self,
        env: Environment,
        capacity: PoolAmount = float('inf'),
        init: PoolAmount = 0,
        hard_cap: bool = False,
        name: Optional[str] = None,
    ):
        self.env = env
        #: Capacity of the pool (maximum level).
        self.capacity = capacity
        #: Current fill level of the pool.
        self.level = init
        self._hard_cap = hard_cap
        self.name = name
        self._put_waiters: List[PoolPutEvent] = []
        self._get_waiters: List[PoolGetEvent] = []
        self._at_most_waiters: List[PoolWhenAtMostEvent] = []
        self._at_least_waiters: List[PoolWhenAtLeastEvent] = []
        self._put_hook: Optional[Callable[[], Any]] = None
        self._get_hook: Optional[Callable[[], Any]] = None
        BoundClass.bind_early(self)

    @property
    def remaining(self) -> PoolAmount:
        """Remaining pool capacity."""
        return self.capacity - self.level

    @property
    def is_empty(self) -> bool:
        """Indicates whether the pool is empty."""
        return self.level == 0

    @property
    def is_full(self) -> bool:
        """Indicates whether the pool is full."""
        return self.level >= self.capacity

    if TYPE_CHECKING:

        def put(self, amount: PoolAmount = 1) -> PoolPutEvent:
            """Put amount in the pool."""
            ...

        def get(self, amount: PoolAmount = 1) -> PoolGetEvent:
            """Get amount from the pool."""
            ...

        def when_at_least(self, amount: PoolAmount) -> PoolWhenAtLeastEvent:
            """Return an event triggered when the pool has at least `amount` items."""
            ...

        def when_at_most(self, amount: PoolAmount) -> PoolWhenAtMostEvent:
            """Return an event triggered when the pool has at most `amount` items."""
            ...

        def when_any(self, epsilon: float = ...) -> PoolWhenAnyEvent:
            """Return an event triggered when the pool is non-empty."""
            ...

        def when_full(self) -> PoolWhenFullEvent:
            """Return an event triggered when the pool becomes full."""
            ...

        def when_not_full(self, epsilon: float = ...) -> PoolWhenNotFullEvent:
            """Return an event triggered when the pool becomes not full."""
            ...

        def when_empty(self) -> PoolWhenEmptyEvent:
            """Return an event triggered when the pool becomes empty."""
            ...

    else:
        put = BoundClass(PoolPutEvent)
        get = BoundClass(PoolGetEvent)
        when_at_least = BoundClass(PoolWhenAtLeastEvent)
        when_at_most = BoundClass(PoolWhenAtMostEvent)
        when_any = BoundClass(PoolWhenAnyEvent)
        when_full = BoundClass(PoolWhenFullEvent)
        when_not_full = BoundClass(PoolWhenNotFullEvent)
        when_empty = BoundClass(PoolWhenEmptyEvent)

    def _trigger_put(self, _: Optional[Event] = None) -> None:
        idx = 0
        while self._put_waiters and idx < len(self._put_waiters):
            put_ev = self._put_waiters[idx]
            if self.capacity - self.level >= put_ev.amount:
                self._put_waiters.pop(idx)
                self.level += put_ev.amount
                put_ev.succeed()
                if self._put_hook:
                    self._put_hook()
            elif self._hard_cap:
                raise OverflowError()
            else:
                idx += 1

    def _trigger_get(self, _: Optional[Event] = None) -> None:
        idx = 0
        while self._get_waiters and idx < len(self._get_waiters):
            get_ev = self._get_waiters[idx]
            if get_ev.amount <= self.level:
                self._get_waiters.pop(idx)
                self.level -= get_ev.amount
                get_ev.succeed(get_ev.amount)
                if self._get_hook:
                    self._get_hook()
            else:
                idx += 1

    def _trigger_when_at_least(self, _: Optional[Event] = None) -> None:
        while self._at_least_waiters and self.level >= self._at_least_waiters[0].amount:
            when_at_least_ev = heappop(self._at_least_waiters)
            when_at_least_ev.succeed()

    def _trigger_when_at_most(self, _: Optional[Event] = None) -> None:
        while self._at_most_waiters and self.level <= self._at_most_waiters[0].amount:
            at_most_ev = heappop(self._at_most_waiters)
            at_most_ev.succeed()

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(name={self.name!r} level={self.level}'
            f' capacity={self.capacity})'
        )


class PriorityPoolPutEvent(Event):
    callbacks: List[EventCallback]

    def __init__(
        self, pool: 'PriorityPool', amount: PoolAmount = 1, priority: int = 0
    ) -> None:
        if not (0 < amount <= pool.capacity):
            raise ValueError('amount must be in (0, capacity]')
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        self.key = priority, pool._event_count
        pool._event_count += 1
        self.callbacks.extend([pool._trigger_when_at_least, pool._trigger_get])
        heappush(pool._put_waiters, self)
        pool._trigger_put()

    def __lt__(self, other: 'PriorityPoolPutEvent') -> bool:
        return self.key < other.key

    def __enter__(self) -> 'PriorityPoolPutEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._put_waiters.remove(self)
            heapify(self.pool._put_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PriorityPoolGetEvent(Event):
    callbacks: List[EventCallback]

    def __init__(self, pool: 'PriorityPool', amount: PoolAmount = 1, priority: int = 0):
        if not (0 < amount <= pool.capacity):
            raise ValueError('amount must be in (0, capacity]')
        super().__init__(pool.env)
        self.pool = pool
        self.amount = amount
        self.key = priority, pool._event_count
        pool._event_count += 1
        self.callbacks.extend([pool._trigger_when_at_most, pool._trigger_put])
        heappush(pool._get_waiters, self)
        pool._trigger_get()

    def __lt__(self, other: 'PriorityPoolGetEvent') -> bool:
        return self.key < other.key

    def __enter__(self) -> 'PriorityPoolGetEvent':
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.cancel()
        return None

    def cancel(self) -> None:
        if not self.triggered:
            self.pool._get_waiters.remove(self)
            heapify(self.pool._get_waiters)
            self.callbacks = None  # type: ignore[assignment] # noqa: F821


class PriorityPool(Pool):
    """Pool with prioritizied put() and get() requests.

    A priority is provided with `put()` and `get()` requests. This priority
    determines the strict order in which requests are fulfilled. Requests of
    the same priority are serviced in strict FIFO order.

    """

    _put_waiters: List[PriorityPoolPutEvent]  # type: ignore[assignment] # noqa: F821
    _get_waiters: List[PriorityPoolGetEvent]  # type: ignore[assignment] # noqa: F821

    def __init__(
        self,
        env: Environment,
        capacity: PoolAmount = float('inf'),
        init: PoolAmount = 0,
        hard_cap: bool = False,
        name: Optional[str] = None,
    ):
        super().__init__(env, capacity, init, hard_cap, name)
        self._event_count = 0

    if TYPE_CHECKING:

        def put(  # type: ignore[override] # noqa: F821
            self, amount: PoolAmount = 1, priority: int = 0
        ) -> PriorityPoolPutEvent:
            """Put amount in the pool."""
            ...

        def get(  # type: ignore[override] # noqa: F821
            self, amount: PoolAmount = 1, priority: int = 0
        ) -> PriorityPoolGetEvent:
            """Get amount from the pool."""
            ...

    else:
        put = BoundClass(PriorityPoolPutEvent)
        get = BoundClass(PriorityPoolGetEvent)

    def _trigger_put(self, _: Optional[Event] = None) -> None:
        while self._put_waiters:
            put_ev = self._put_waiters[0]
            if self.capacity - self.level >= put_ev.amount:
                heappop(self._put_waiters)
                self.level += put_ev.amount
                put_ev.succeed()
                if self._put_hook:
                    self._put_hook()
            elif self._hard_cap:
                raise OverflowError()
            else:
                break

    def _trigger_get(self, _: Optional[Event] = None) -> None:
        while self._get_waiters:
            get_ev = self._get_waiters[0]
            if get_ev.amount <= self.level:
                heappop(self._get_waiters)
                self.level -= get_ev.amount
                get_ev.succeed(get_ev.amount)
                if self._get_hook:
                    self._get_hook()
            else:
                break



"""Tools for managing simulation configurations.

Each simulation requires a configuration dictionary that defines various
configuration values for both the simulation (`desmod`) and the user model. The
configuration dictionary is flat, but the keys use a dotted notation, similar
to :class:`Component` scopes, that allows for different namespaces to exist
within the [flat] configuration dictionary.

Several configuration key/values are required by desmod itself. These
configuration keys are prefixed with 'sim.'; for example: 'sim.duration' and
'sim.seed'.

Models may define their own configuration key/values, but should avoid using
the 'sim.` prefix.

The :class:`NamedManager` class provides a mechanism for defining named
groupings of configuration values. These named configuration groups allow
quick configuration of multiple values. Configuration groups are also
composable: a configuration group can be defined to depend on several other
configuration groups.

Most functions in this module are provided to support building user interfaces
for configuring a model.

"""
from collections.abc import Sequence
from copy import deepcopy
from itertools import product
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
)
import builtins

ConfigDict = Dict[str, Any]
ConfigFactor = Tuple[List[str], List[Any]]


class ConfigError(Exception):
    """Exception raised for a variety of configuration errors."""


class NamedConfig(NamedTuple):
    """Named configuration group details.

    Iterating a :class:`NamedManager` instance yields ``NamedConfig``
    instances.

    """

    category: str
    name: str
    doc: str
    depend: List[str]
    config: ConfigDict


class NamedManager:
    """Manage named configuration groups.

    Any number of named configuration groups can be specified using the
    :meth:`name()` method. The :meth:`resolve()` method is used to compose a
    fully-resolved configuration based on one or more configuration group
    names.

    Iterating a ``NamedManager`` instance will yield :class:`NamedConfig`
    instances for each registered named configuration.

    """

    def __init__(self) -> None:
        self._named_configs: Dict[str, NamedConfig] = {}

    def name(
        self,
        name: str,
        depend: Optional[List[str]] = None,
        config: Optional[ConfigDict] = None,
        category: str = '',
        doc: str = '',
    ) -> None:
        """Declare a new configuration group.

        A configuration group consists of a name, a list of dependencies, and a
        dictionary of configuration key/values. This function declares a new
        configuration group that may be later resolved with :meth:`resolve()`.

        :param str name: Name of new configuration group.
        :param list depend: List of configuration group dependencies.
        :param dict config: Configuration key/values.
        :param str category: Optional category.
        :param str doc: Optional documentation for named configuration group.

        """
        if name in self._named_configs:
            raise ConfigError(f'name already used: {name}')
        if depend is None:
            depend = []
        if config is None:
            config = {}
        self._named_configs[name] = NamedConfig(category, name, doc, depend, config)

    def resolve(self, *names: str) -> ConfigDict:
        """Resolve named configs into a new config object."""
        resolved: ConfigDict = {}
        self._resolve(resolved, *names)
        return resolved

    def _resolve(self, resolved: ConfigDict, *names: str) -> None:
        for name in names:
            if name not in self._named_configs:
                raise ConfigError(f'unknown named config: {name}')
            nc = self._named_configs[name]
            self._resolve(resolved, *nc.depend)
            resolved.update(nc.config)

    def __iter__(self) -> Iterator[NamedConfig]:
        """Iterate named config tuples."""
        yield from self._named_configs.values()


def apply_user_config(config: ConfigDict, user_config: ConfigDict) -> None:
    """Apply user-provided configuration to a configuration.

    Each key/value from `user_config` is validated and then used to override
    the same key in `config`.

    :param dict config: The configuration to update.
    :param dict user_config: The user-provided config with overriding items.
    :raises .ConfigError: For invalid user keys or values.

    """
    for key, value in user_config.items():
        try:
            current_value = config[key]
        except KeyError:
            raise ConfigError(f'Invalid config key: {key}')
        current_type = type(current_value)
        if not (
            isinstance(value, current_type)
            # allow new float value to replace integer default without truncation from
            # int to float coercion
            or (isinstance(value, float) and issubclass(current_type, int))
        ):
            try:
                value = current_type(value)
            except (ValueError, TypeError):
                raise ConfigError(
                    f'Failed to coerce {value} to {current_type.__name__} for {key}'
                )
        config[key] = value


def apply_user_overrides(
    config: ConfigDict,
    overrides: Iterable[Tuple[str, str]],
    eval_locals: Optional[Dict[str, Any]] = None,
) -> None:
    """Apply user-provided overrides to a configuration.

    The user-provided `overrides` list are first verified for validity and
    then applied to the the provided `config` dictionary.

    Each user-provided key must already exist in `config`. The
    :func:`fuzzy_lookup()` function is used to verify that the user-provided
    key exists unambiguously in `config`.

    The user-provided value expressions are evaluated against a safe local
    environment using :func:`eval()`. The type of the resulting value must be
    type-compatible with the existing (default) value in `config`.

    :param dict config: Configuration dictionary to modify.
    :param list overrides:
        List of user-provided (key, value expression) tuples.
    :param dict eval_locals:
        Optional dictionary of locals to use with :func:`eval()`. A safe and
        useful set of locals is provided by default.

    """
    for user_key, user_expr in overrides:
        key, current_value = fuzzy_lookup(config, user_key)
        value = _safe_eval(user_expr, type(current_value), eval_locals)
        config[key] = value


def parse_user_factors(
    config: ConfigDict, user_factors, eval_locals: Optional[Dict[str, Any]] = None
) -> List[ConfigFactor]:
    """Safely parse user-provided configuration factors.

    A configuration factor consists of an n-tuple of configuration keys along
    with a list of corresponding n-tuples of values. Configuration factors are
    used by :func:`~desmod.simulation.simulate_factors()` to run multiple
    simulations to explore a subset of the model's configuration space.

    :param dict config:
        The configuration dictionary is used to check the keys and values of
        the user-provided factors. The dictionary is not modified.
    :param user_factors:
        Sequence of `(user_keys, user_expressions)` tuples. See
        :func:`parse_user_factor()` for more detail on user keys and
        expressions.
    :param dict eval_locals:
        Optional dictionary of locals used when :func:`eval()`-ing user
        expressions.
    :returns:
        List of keys, values pairs. The returned list of factors is suitable
        for passing to :func:`simulate_factors`.
    :raises .ConfigError: For invalid user keys or expressions.

    """
    return [
        parse_user_factor(config, user_keys, user_exprs, eval_locals)
        for user_keys, user_exprs in user_factors
    ]


def parse_user_factor(
    config: ConfigDict,
    user_keys: str,
    user_exprs: str,
    eval_locals: Optional[Dict[str, Any]] = None,
) -> ConfigFactor:
    """Safely parse a user-provided configuration factor.

    Example:

        >>> config = {'a.b.x': 0,
                      'a.b.y': True,
                      'a.b.z': 'something'}
        >>> parse_user_factor(config, 'x,y', '(1,True), (2,False), (3,True)')
        [['a.b.x', 'a.b.y'], [[1, True], [2, False], [3, True]]]

    :param dict config:
        The configuration dictionary is used to check the keys and values of
        the user-provided factors. The dictionary is not modified.
    :param str user_keys:
        String of comma-separated configuration keys of the factor. The keys
        may be fuzzy (i.e. valid for use with :func:`fuzzy_lookup()`), but note
        that the returned keys will always be fully-qualified (non-fuzzy).
    :param str user_exprs:
        User-provided Python expressions string. The expressions string is
        evaluated using :func:`eval()` with, by default, a safe locals
        dictionary. The expressions string must evaluate to a sequence of
        n-tuples where `n` is the number of keys provided in `user_keys`.
        Further, the elements of each n-tuple must be type-compatible with the
        existing (default) values in the `config` dict.
    :param dict eval_locals:
        Optional dictionary of locals used when :func:`eval()`-ing user
        expressions.

    :returns:
        A config factor: a pair (2-list) of keys and values lists.

        .. Note:: All sequences in the returned factor are expressed as lists,
                  not tuples. This is done to improve YAML serialization.

    :raises .ConfigError: For invalid keys or value expressions.

    """
    current = [
        fuzzy_lookup(config, user_key.strip()) for user_key in user_keys.split(',')
    ]
    user_values = _safe_eval(user_exprs, eval_locals=eval_locals)
    values = []
    if not isinstance(user_values, Sequence):
        raise ConfigError(f'Factor value not a sequence "{user_values}"')
    for user_items in user_values:
        if len(current) == 1:
            user_items = [user_items]
        items = []
        for (key, current_value), item in zip(current, user_items):
            current_type = type(current_value)
            if not isinstance(item, current_type):
                try:
                    item = current_type(item)
                except (ValueError, TypeError):
                    raise ConfigError(
                        f'Failed to coerce {item} to {current_type.__name__}'
                    )
            items.append(item)
        values.append(items)
    return ([key for key, _ in current], values)


def factorial_config(
    base_config: ConfigDict,
    factors: Iterable[ConfigFactor],
    special_key: Optional[str] = None,
) -> Iterator[ConfigDict]:
    """Generate configurations from base config and config factors.

    :param dict base_config:
        Configuration dictionary that the generated configuration dictionaries
        are based on. This dict is not modified; generated config dicts are
        created with :func:`copy.deepcopy()`.
    :param list factors:
        Sequence of one or more configuration factors. Each configuration
        factor is a 2-tuple of keys and values lists.
    :param str special_key:
        When specified, a key/value will be inserted into the generated
        configuration dicts that identifies the "special" (unique) key/value
        combinations of the specified `factors` used in the config dict.
    :yields:
        Configuration dictionaries with the cartesian product of the provided
        `factors` applied. I.e. each yielded config dict will have a unique
        combination of the `factors`.

    """
    unrolled_factors = []
    for keys, values_list in factors:
        unrolled_factors.append([(keys, values) for values in values_list])

    for keys_values_lists in product(*unrolled_factors):
        config = deepcopy(base_config)
        special: List[Tuple[str, Any]] = []
        if special_key:
            config[special_key] = special
        for keys, values in keys_values_lists:
            for key, value in zip(keys, values):
                config[key] = value
                if special_key:
                    special.append((key, value))
        yield config


def fuzzy_match(keys: Iterable[str], fuzzy_key: str) -> str:
    """Match a fuzzy key against sequence of canonical key names.

    :param keys: Sequence of canonical key names.
    :param str fuzzy_key: Fuzzy key to match against canonical keys.
    :returns: Canonical matching key name.
    :raises KeyError: If fuzzy key does not match.

    """
    suffix_matches = []
    split_matches = []
    for k in keys:
        if k == fuzzy_key:
            return k
        elif k.rsplit('.', 1)[-1] == fuzzy_key:
            split_matches.append(k)
        elif k.endswith(fuzzy_key):
            suffix_matches.append(k)
    if len(split_matches) == 1:
        return split_matches[0]
    elif len(suffix_matches) == 1:
        return suffix_matches[0]
    elif not any((suffix_matches, split_matches)):
        raise KeyError(fuzzy_key)
    else:
        raise KeyError(fuzzy_key + ' is ambiguous')


def fuzzy_lookup(config: ConfigDict, fuzzy_key: str) -> Tuple[str, Any]:
    """Lookup a config key/value using a partially specified (fuzzy) key.

    The lookup will succeed iff the provided `fuzzy_key` unambiguously matches
    the tail of a [fully-qualified] key in the `config` dict.

    :param dict config: Configuration dict in which to lookup `fuzzy_key`.
    :param str fuzzy_key: Partially specified key to lookup in `config`.
    :returns:
        `(key, value)` tuple. The returned key is the regular, fully-qualified
        key name, not the provided `fuzzy_key`.
    :raises .ConfigError: For non-matching `fuzzy_key`.

    """
    try:
        k = fuzzy_match(config, fuzzy_key)
    except KeyError as e:
        raise ConfigError(f'Invalid config key: {e}')
    else:
        return k, config[k]


_safe_builtins = [
    'abs',
    'bin',
    'bool',
    'dict',
    'float',
    'frozenset',
    'hex',
    'int',
    'len',
    'list',
    'max',
    'min',
    'oct',
    'ord',
    'range',
    'round',
    'set',
    'str',
    'sum',
    'tuple',
    'tuple',
    'zip',
    'True',
    'False',
]

_default_eval_locals = {
    name: getattr(builtins, name) for name in _safe_builtins if hasattr(builtins, name)
}


def _safe_eval(
    expr: str,
    coerce_type: Optional[Type] = None,
    eval_locals: Optional[Dict[str, Any]] = None,
) -> Any:
    if eval_locals is None:
        eval_locals = _default_eval_locals
    try:
        value = eval(expr, {'__builtins__': None}, eval_locals)
    except BaseException:
        if coerce_type and issubclass(coerce_type, str):
            value = expr
        else:
            raise ConfigError(f'Failed evaluation of expression "{expr}"')

    if coerce_type:
        if expr in eval_locals and not isinstance(value, coerce_type):
            value = expr
        if not isinstance(value, coerce_type):
            try:
                value = coerce_type(value)
            except (ValueError, TypeError):
                raise ConfigError(
                    f'Failed to coerce expression {_quote_expr(expr)} to '
                    f'{coerce_type.__name__}'
                )
    return value


def _quote_expr(expr: str) -> str:
    quote_char = "'" if expr.startswith('"') else '"'
    return ''.join([quote_char, expr, quote_char])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools_scm import get_version

# flake8: noqa

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
]

autodoc_member_order = 'bysource'
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'desmod'
copyright = '2016, Western Digital Corporation'
author = 'Pete Grayson'
version = get_version(root='..', relative_to=__file__)
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'sphinx'
todo_include_todos = True

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
htmlhelp_basename = 'desmoddoc'

latex_elements = {}
latex_documents = [
    (master_doc, 'desmod.tex', 'desmod Documentation', 'Pete Grayson', 'manual'),
]
man_pages = [(master_doc, 'desmod', 'desmod Documentation', [author], 1)]
texinfo_documents = [
    (
        master_doc,
        'desmod',
        'desmod Documentation',
        author,
        'desmod',
        'Discrete Event Simulation Modeling using SimPy',
        'Miscellaneous',
    ),
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'simpy': ('https://simpy.readthedocs.io/en/latest/', None),
}



"""Model grocery store checkout lanes.

A grocery store checkout system is modeled. Each grocery store has one or more
checkout lanes. Each lane has a cashier that scans customers' items. Zero or
more baggers bag items after the cashier scans them. Cashiers will also bag
items if there is no bagger helping at their lane.

Several bagger assignment policies are implemented. This model helps determine
the optimal policy under various conditions. The model is also useful for
estimating bagger, checkout lane, and cashier resources needed for various
customer profiles.

"""
from argparse import ArgumentParser
from datetime import timedelta
from functools import partial
from itertools import count

from simpy import Container, Resource
from vcd.gtkw import GTKWSave

from desmod.component import Component
from desmod.config import apply_user_overrides, parse_user_factors
from desmod.dot import generate_dot
from desmod.queue import Queue
from desmod.simulation import simulate, simulate_factors


class Top(Component):
    """The top-level component of the model."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.customers = Customers(self)
        self.grocery = GroceryStore(self)

    def connect_children(self):
        self.connect(self.customers, 'grocery')

    @classmethod
    def pre_init(cls, env):
        # Compose a GTKWave save file that lays-out the various VCD signals in
        # a meaningful manner. This must be done at pre-init time to allow
        # sim.gtkw.live to work.
        analog_kwargs = {
            'datafmt': 'dec',
            'color': 'cycle',
            'extraflags': ['analog_step'],
        }
        with open(env.config['sim.gtkw.file'], 'w') as gtkw_file:
            gtkw = GTKWSave(gtkw_file)
            gtkw.dumpfile(env.config['sim.vcd.dump_file'], abspath=False)
            gtkw.treeopen('grocery')
            gtkw.signals_width(300)
            gtkw.trace('customers.active', **analog_kwargs)
            for i in range(env.config['grocery.num_lanes']):
                with gtkw.group(f'Lane{i}'):
                    scope = f'grocery.lane{i}'
                    gtkw.trace(f'{scope}.customer_queue', **analog_kwargs)
                    gtkw.trace(f'{scope}.feed_belt', **analog_kwargs)
                    gtkw.trace(f'{scope}.bag_area', **analog_kwargs)
                    gtkw.trace(f'{scope}.baggers', **analog_kwargs)

    def elab_hook(self):
        # We generate DOT representations of the component hierarchy. It is
        # only after elaboration that the component tree is fully populated and
        # connected, thus generate_dot() is called here in elab_hook().
        generate_dot(self)


class GroceryStore(Component):
    """Model a grocery store with checkout lanes, cashiers, and baggers."""

    base_name = 'grocery'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        num_lanes = self.env.config['grocery.num_lanes']
        self.checkout_lanes = [CheckoutLane(self, index=i) for i in range(num_lanes)]

        num_baggers = self.env.config['grocery.num_baggers']
        self.baggers = [Bagger(self, index=i) for i in range(num_baggers)]

    def connect_children(self):
        # The baggers move between checkout lanes depending on bagger.policy,
        # so each bagger must be connected to all of the checkout lanes.
        for bagger in self.baggers:
            self.connect(bagger, 'checkout_lanes')


class CheckoutLane(Component):
    """Model a grocery store checkout lane.

    Each lane has a customer queue which is modeled with a
    :class:`simpy.Resource`. Customers are addressed in a first-come,
    first-serve manner.

    Once a customer reaches the front of the checkout lane's line, they place
    their items on the lane's feed belt. The feed belt is modeled as a
    :class:`desmod.queue.Queue` and has limited capacity for items. That
    capacity is configurable with `checkout.feed_capacity`.

    The checkout lane's cashier takes items from the feed belt, scans them, and
    places them in the lane's bagging area. The bagging area is also modeled as
    a queue with limited capacity.

    Finally, a checkout lane may have zero or more baggers assigned to it. A
    :class:`simpy.Container` is used to keep track of how many baggers are
    present at each lane.

    """

    base_name = 'lane'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cashier = Cashier(self)

        self.customer_queue = Resource(self.env)
        self.auto_probe('customer_queue', trace_queue=True, vcd={'init': 0})

        feed_capacity = self.env.config['checkout.feed_capacity']
        self.feed_belt = Queue(self.env, capacity=feed_capacity)
        self.auto_probe('feed_belt', vcd={})

        bag_area_capacity = self.env.config['checkout.bag_area_capacity']
        self.bag_area = Queue(self.env, capacity=bag_area_capacity)
        self.auto_probe('bag_area', vcd={})

        self.baggers = Container(self.env)
        self.auto_probe('baggers', vcd={})

    def connect_children(self):
        self.connect(self.cashier, 'lane', conn_obj=self)


class Cashier(Component):
    """Model checkout lane cashier.

    A cashier occupies a single checkout lane. They take items from the lane's
    feeder belt, scan the item, and place the item in the lane's bagging area.

    If no bagger personel is present, the cashier will also perform bagging,
    but a cashier cannot scan and bag at the same time.

    """

    base_name = 'cashier'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_connections('lane')
        self.add_process(self.checkout)

        # Use exponential distribution to model item scan and bag times.
        self.scan_dist = partial(
            self.env.rand.expovariate, 1 / self.env.config['cashier.scan_time']
        )

        self.bag_dist = partial(
            self.env.rand.expovariate, 1 / self.env.config['cashier.bag_time']
        )

    def checkout(self):
        """Cashier checkout behavior."""
        while True:
            if not self.lane.baggers.level and self.lane.bag_area.is_full:
                yield self.env.process(self.bag_items())
            item, done_event = yield self.lane.feed_belt.get()
            yield self.env.timeout(self.scan_dist())
            yield self.lane.bag_area.put((item, done_event))
            if done_event is not None:
                # A customer's final item comes with a done event. The bag area
                # must be emptied between customers.
                yield self.env.process(self.bag_items())

    def bag_items(self):
        while not self.lane.bag_area.is_empty and not self.lane.baggers.level:
            item, done_event = yield self.lane.bag_area.get()
            yield self.env.timeout(self.bag_dist())
            if done_event is not None:
                # Notify the customer that their checkout is complete.
                done_event.succeed()


class Bagger(Component):
    """Model a grocery store bagger employee.

    Baggers take customer items from checkout lane bagging areas and bag them.

    Several policies for how baggers are assigned to checkout lanes can be
    configured with `bagger.policy`.

    """

    base_name = 'bagger'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_connections('checkout_lanes')
        self.bag_dist = partial(
            self.env.rand.expovariate, 1 / self.env.config['bagger.bag_time']
        )

        policy = self.env.config['bagger.policy']
        if policy == 'float-aggressive':
            self.add_process(self.policy_float_aggressive)
        elif policy == 'float-lazy':
            self.add_process(self.policy_float_lazy)
        elif policy == 'fixed-lane':
            self.add_process(self.policy_fixed_lane)
        else:
            raise ValueError(f'invalid bagger.policy {policy}')

    def policy_float_aggressive(self):
        """Assign bagger to the first lane with any baggable items.

        The bagger floats between checkout lanes. As soon as a lane is
        identified with any baggable items, the bagger assigns to that lane and
        bags until the lane's bag area is empty.

        """
        while True:
            yield self.env.any_of(
                lane.bag_area.when_any() for lane in self.checkout_lanes
            )

            lanes = reversed(
                sorted(
                    filter(lambda lane: not lane.baggers.level, self.checkout_lanes),
                    key=lambda lane: lane.bag_area.size,
                )
            )
            for lane in lanes:
                yield lane.baggers.put(1)
                self.debug('assigned to lane', lane.index)
                yield self.env.process(self.bag_items(lane.bag_area))
                yield lane.baggers.get(1)
                self.debug('leave lane', lane.index)
                break

    def policy_float_lazy(self):
        """Assign bagger to lane with full bagging area.

        The bagger remains idle until he identifies a lane with a full bagging
        area. The bagger bags at that lane until the bagging area is emptied.

        """
        while True:
            yield self.env.any_of(
                lane.bag_area.when_full() for lane in self.checkout_lanes
            )

            for lane in filter(lambda lane: lane.bag_area.is_full, self.checkout_lanes):
                yield lane.baggers.put(1)
                self.debug('assigned to lane', lane.index)
                yield self.env.process(self.bag_items(lane.bag_area))
                yield lane.baggers.get(1)
                self.debug('leave lane', lane.index)
                break

    def policy_fixed_lane(self):
        """Static assignment of bagger to a lane.

        The bagger finds the first lane with no other baggers and stays there.

        """
        _, lane = min((lane.baggers.level, lane) for lane in self.checkout_lanes)
        yield lane.baggers.put(1)
        self.debug('assigned to lane', lane.index)
        while True:
            yield lane.bag_area.when_any()
            yield self.env.process(self.bag_items(lane.bag_area))

    def bag_items(self, bag_area):
        while not bag_area.is_empty:
            item, done_event = yield bag_area.get()
            yield self.env.timeout(self.bag_dist())
            if done_event is not None:
                done_event.succeed()


class Customers(Component):
    """Model customer arrival rate and in-store behavior.

    Each customer's arrival time, number of items, and shopping time is
    determined by configuration.

    A new process is spawned for each customer.

    A "customers" database table captures per-customer checkout times. A
    primary goal for this model is optimizing customer checkout time (latency)
    and throughput.

    """

    base_name = 'customers'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_connections('grocery')
        self.add_process(self.generate_customers)
        self.active = Container(self.env)
        self.auto_probe('active', vcd={})
        if self.env.tracemgr.sqlite_tracer.enabled:
            self.db = self.env.tracemgr.sqlite_tracer.db
            self.db.execute(
                'CREATE TABLE customers '
                '(cust_id INTEGER PRIMARY KEY,'
                ' num_items INTEGER,'
                ' shop_time REAL,'
                ' checkout_time REAL)'
            )
        else:
            self.db = None

    def generate_customers(self):
        """Generate grocery store customers.

        Various configuration parameters determine the distribution of customer
        arrival times as well as the number of items each customer will shop
        for.

        """
        cust_id = count()
        arrival_interval_dist = partial(
            self.env.rand.expovariate, 1 / self.env.config['customer.arrival_interval']
        )
        time_per_item_dist = partial(
            self.env.rand.expovariate, 1 / self.env.config['customer.time_per_item']
        )
        num_items_mu = self.env.config['customer.num_items.mu']
        num_items_sigma = self.env.config['customer.num_items.sigma']
        num_items_dist = partial(
            self.env.rand.normalvariate, num_items_mu, num_items_sigma
        )

        while True:
            num_items = max(1, round(num_items_dist()))
            self.env.process(
                self.customer(
                    next(cust_id), num_items, shop_time=num_items * time_per_item_dist()
                )
            )
            yield self.env.timeout(arrival_interval_dist())

    def customer(self, cust_id, num_items, shop_time):
        """Grocery store customer behavior."""
        yield self.active.put(1)
        self.debug(cust_id, 'start shopping for', num_items, 'items')
        yield self.env.timeout(shop_time)
        self.debug(cust_id, 'ready to checkout after', timedelta(seconds=shop_time))

        t0 = self.env.now

        lane = sorted(
            self.grocery.checkout_lanes, key=lambda lane: len(lane.customer_queue.queue)
        )[0]

        with lane.customer_queue.request() as req:
            self.debug('enter queue', lane.index)
            yield req
            for i in range(num_items - 1):
                yield lane.feed_belt.put((i, None))
            checkout_done = self.env.event()
            yield lane.feed_belt.put((num_items - 1, checkout_done))

        yield checkout_done
        checkout_time = self.env.now - t0
        self.debug(cust_id, 'done checking out after', timedelta(seconds=checkout_time))
        yield self.active.get(1)
        if self.db:
            self.db.execute(
                'INSERT INTO customers '
                '(cust_id, num_items, shop_time, checkout_time) '
                'VALUES (?,?,?,?)',
                (cust_id, num_items, shop_time, checkout_time),
            )

    def get_result_hook(self, result):
        if not self.db:
            return
        result['checkout_time_avg'] = self.db.execute(
            'SELECT AVG(checkout_time) FROM customers'
        ).fetchone()[0]
        result['checkout_time_min'] = self.db.execute(
            'SELECT MIN(checkout_time) FROM customers'
        ).fetchone()[0]
        result['checkout_time_max'] = self.db.execute(
            'SELECT MAX(checkout_time) FROM customers'
        ).fetchone()[0]
        result['customers_total'] = self.db.execute(
            'SELECT COUNT() FROM customers'
        ).fetchone()[0]
        result['customers_per_hour'] = result['customers_total'] / (
            self.env.time() / 3600
        )


if __name__ == '__main__':
    config = {
        'bagger.bag_time': 1.5,
        'bagger.policy': 'float-aggressive',
        'cashier.bag_time': 2.0,
        'cashier.scan_time': 2.0,
        'checkout.bag_area_capacity': 15,
        'checkout.feed_capacity': 20,
        'customer.arrival_interval': 60,
        'customer.num_items.mu': 50,
        'customer.num_items.sigma': 10,
        'customer.time_per_item': 30.0,
        'grocery.num_baggers': 1,
        'grocery.num_lanes': 2,
        'sim.db.enable': True,
        'sim.db.persist': False,
        'sim.dot.colorscheme': 'blues5',
        'sim.dot.enable': True,
        'sim.duration': '7200 s',
        'sim.gtkw.file': 'sim.gtkw',
        'sim.gtkw.live': False,
        'sim.log.enable': True,
        'sim.progress.enable': False,
        'sim.result.file': 'result.json',
        'sim.seed': 1234,
        'sim.timescale': 's',
        'sim.vcd.dump_file': 'sim.vcd',
        'sim.vcd.enable': True,
        'sim.vcd.persist': False,
        'sim.workspace': 'workspace',
    }

    parser = ArgumentParser()
    parser.add_argument(
        '--set',
        '-s',
        nargs=2,
        metavar=('KEY', 'VALUE'),
        action='append',
        default=[],
        dest='config_overrides',
        help='Override config KEY with VALUE expression',
    )
    parser.add_argument(
        '--factor',
        '-f',
        nargs=2,
        metavar=('KEYS', 'VALUES'),
        action='append',
        default=[],
        dest='factors',
        help='Add multi-factor VALUES for KEY(S)',
    )
    args = parser.parse_args()
    apply_user_overrides(config, args.config_overrides)
    factors = parse_user_factors(config, args.factors)
    if factors:
        simulate_factors(config, factors, Top)
    else:
        simulate(config, Top)



import os
import sqlite3

import pytest
import simpy

from desmod.component import Component
from desmod.pool import Pool
from desmod.queue import Queue
from desmod.simulation import simulate

pytestmark = pytest.mark.usefixtures('cleandir')


@pytest.fixture
def cleandir(tmpdir):
    origin = os.getcwd()
    tmpdir.chdir()
    yield None
    os.chdir(origin)


@pytest.fixture
def config():
    return {
        'sim.db.enable': False,
        'sim.db.file': 'sim.sqlite',
        'sim.duration': '10 us',
        'sim.log.enable': False,
        'sim.log.file': 'sim.log',
        'sim.log.level': 'INFO',
        'sim.result.file': 'result.yaml',
        'sim.seed': 1234,
        'sim.timescale': '1 us',
        'sim.vcd.dump_file': 'sim.vcd',
        'sim.vcd.enable': False,
        'sim.vcd.gtkw_file': 'sim.gtkw',
        'sim.vcd.start_time': '',
        'sim.vcd.stop_time': '',
        'sim.workspace': 'workspace',
        'test.raise': False,
    }


class TopTest(Component):

    base_name = 'top'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.container = simpy.Container(self.env)
        self.resource = simpy.Resource(self.env)
        self.queue = Queue(self.env)
        self.pool = Pool(self.env)
        self.a = CompA(self)
        self.b = CompB(self)
        hints = {}
        if self.env.config['sim.log.enable']:
            hints['log'] = {'level': 'INFO'}
        if self.env.config['sim.vcd.enable']:
            hints['vcd'] = {}
        if self.env.config['sim.db.enable']:
            hints['db'] = {}
        self.auto_probe('container', **hints)
        self.auto_probe('resource', **hints)
        self.auto_probe('queue', **hints)
        self.auto_probe('pool', **hints)
        self.trace_some = self.get_trace_function(
            'something', vcd={'var_type': 'real'}, log={'level': 'INFO'}
        )
        self.trace_other = self.get_trace_function(
            'otherthing',
            vcd={'var_type': 'integer', 'init': ('z', 'z'), 'size': (8, 8)},
        )
        self.add_process(self.loop)

    def connect_children(self):
        self.connect(self.a, 'container')
        self.connect(self.b, 'container')

    def loop(self):
        while True:
            yield self.env.timeout(5)
            with self.resource.request() as req:
                yield req
            self.trace_some(17.0)
            self.trace_other(42, 17)
            if self.env.config.get('test.raise'):
                raise Exception('oops')


class CompA(Component):

    base_name = 'a'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_process(self.loop)
        self.add_connections('container')

    def loop(self):
        while True:
            yield self.container.get(3)


class CompB(CompA):

    base_name = 'b'

    def loop(self):
        while True:
            yield self.container.put(1)
            yield self.env.timeout(1)


def test_defaults(config):
    simulate(config, TopTest)
    workspace = config['sim.workspace']
    assert os.path.isdir(workspace)
    assert os.path.exists(os.path.join(workspace, config['sim.result.file']))
    for filename_key in [
        'sim.log.file',
        'sim.vcd.dump_file',
        'sim.vcd.gtkw_file',
        'sim.db.file',
    ]:
        assert not os.path.exists(os.path.join(workspace, config[filename_key]))


def test_exception(config):
    config['sim.log.enable'] = True
    config['test.raise'] = True
    with pytest.raises(Exception):
        simulate(config, TopTest)
    log_path = os.path.join(config['sim.workspace'], config['sim.log.file'])
    assert os.path.exists(log_path)
    with open(log_path) as f:
        log = f.read()
    assert 'ERROR' in log


def test_log(config):
    config['sim.log.enable'] = True
    simulate(config, TopTest)
    log_path = os.path.join(config['sim.workspace'], config['sim.log.file'])
    assert os.path.exists(log_path)
    last_line = open(log_path).readlines()[-1]
    assert last_line == 'INFO    9.000 us: top.container: 1\n'


def test_log_stderr(config, capsys):
    config['sim.log.enable'] = True
    config['sim.log.file'] = ''
    simulate(config, TopTest)
    out, err = capsys.readouterr()
    assert out == ''
    assert err.endswith('INFO    9.000 us: top.container: 1\n')


def test_log_persist(config):
    config['sim.log.enable'] = True
    config['sim.log.persist'] = False
    simulate(config, TopTest)
    log_path = os.path.join(config['sim.workspace'], config['sim.log.file'])
    assert not os.path.exists(log_path)

    config['sim.log.file'] = ''
    simulate(config, TopTest)


def test_vcd(config):
    config['sim.vcd.enable'] = True
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    assert os.path.exists(dump_path)
    with open(dump_path) as dump:
        vcd_str = dump.read()
        for t in range(1, 11):
            assert f'#{t}\n' in vcd_str


def test_vcd_start(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.start_time'] = '5 us'
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    with open(dump_path) as dump:
        vcd_str = dump.read()
        assert 'dumpon' in vcd_str
        assert '#6' in vcd_str


def test_vcd_stop(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.stop_time'] = '5 us'
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    with open(dump_path) as dump:
        vcd_str = dump.read()
        assert 'dumpoff' in vcd_str
        assert '#6' not in vcd_str


def test_vcd_start_then_stop(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.start_time'] = '4 us'
    config['sim.vcd.stop_time'] = '6 us'
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    with open(dump_path) as dump:
        vcd_str = dump.read()
        assert 'dumpon' in vcd_str
        assert 'dumpoff' in vcd_str
        assert '#1\n' not in vcd_str
        assert '#5' in vcd_str
        assert '#9' not in vcd_str


def test_vcd_stop_then_start(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.start_time'] = '6 us'
    config['sim.vcd.stop_time'] = '4 us'
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    with open(dump_path) as dump:
        vcd_str = dump.read()
        assert 'dumpon' in vcd_str
        assert 'dumpoff' in vcd_str
        assert '#1\n' in vcd_str
        assert '#5' not in vcd_str
        assert '#9' in vcd_str


def test_vcd_timescale(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.timescale'] = '10 s'
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    with open(dump_path) as dump:
        vcd_str = dump.read()
        assert '$timescale 10 s' in vcd_str


def test_vcd_persist(config):
    config['sim.vcd.enable'] = True
    config['sim.vcd.persist'] = False
    simulate(config, TopTest)
    dump_path = os.path.join(config['sim.workspace'], config['sim.vcd.dump_file'])
    assert not os.path.exists(dump_path)


def test_db(config):
    config['sim.db.enable'] = True
    simulate(config, TopTest)
    db_path = os.path.join(config['sim.workspace'], config['sim.db.file'])
    assert os.path.exists(db_path)
    db = sqlite3.connect(db_path)
    assert db.execute('SELECT COUNT() FROM trace').fetchone()[0] == 15


def test_db_persist(config):
    config['sim.db.enable'] = True
    config['sim.db.persist'] = False
    simulate(config, TopTest)
    db_path = os.path.join(config['sim.workspace'], config['sim.db.file'])
    assert not os.path.exists(db_path)


def test_db_include_pat(config):
    config['sim.db.enable'] = True
    config['sim.db.include_pat'] = [r'top\.resource']
    simulate(config, TopTest)
    db_path = os.path.join(config['sim.workspace'], config['sim.db.file'])
    assert os.path.exists(db_path)
    db = sqlite3.connect(db_path)
    assert db.execute('SELECT COUNT() FROM trace').fetchone()[0] == 2


def test_db_in_memory(config):
    config['sim.db.enable'] = True
    config['sim.db.file'] = ':memory:'
    simulate(config, TopTest)
    db_path = os.path.join(config['sim.workspace'], config['sim.db.file'])
    assert not os.path.exists(db_path)



import sys

import pytest

from desmod.config import (
    ConfigError,
    NamedManager,
    _safe_eval,
    apply_user_config,
    apply_user_overrides,
    factorial_config,
    fuzzy_lookup,
    parse_user_factor,
    parse_user_factors,
)


@pytest.fixture
def config():
    return {
        'foo.bar.baz': 17,
        'foo.bar.biz': 1.23,
        'abc.def.baz': False,
        'a.b.c': 'something',
        'd.e.f': [3, 2, 1],
        'g.h.i': {'a': 1, 'b': 2},
    }


@pytest.fixture
def named_mgr():
    return NamedManager()


def test_named_reuse(named_mgr):
    named_mgr.name('xxx', [], {'x': 0})
    with pytest.raises(ConfigError):
        named_mgr.name('xxx', [], {'y': 1})
    with pytest.raises(ConfigError):
        named_mgr.resolve('yyy')


def test_named_resolve(named_mgr):
    named_mgr.name('www', config={'w': 0})
    named_mgr.name('xxx', [], {'x': 1}, category='thing', doc='documentation')
    named_mgr.name('yyy', ['xxx', 'www'], {'y': 2})
    named_mgr.name('zzz', depend=['yyy'], config={'z': 3})
    named_mgr.name('qqq', ['zzz'])
    assert named_mgr.resolve('qqq') == {'w': 0, 'x': 1, 'y': 2, 'z': 3}

    assert set(nc.name for nc in named_mgr) == {'www', 'xxx', 'yyy', 'zzz', 'qqq'}
    for nc in named_mgr:
        if nc.name == 'xxx':
            assert nc.category == 'thing' and nc.doc == 'documentation'
        else:
            assert not nc.category and not nc.doc


@pytest.mark.parametrize(
    'fuzzy_key, expected',
    [
        ('foo', ConfigError),
        ('b.foo', ('a.b.foo', 1)),
        ('d.foo', ('c.d.foo', 3)),
        ('bar', ('a.b.bar', 2)),
        ('o', ('e.f.o', 4)),
        ('.o', ('e.f.o', 4)),
        ('x.y.z', ('x.y.z', 5)),
        ('y.z', ConfigError),
    ],
)
def test_fuzzy_lookup(fuzzy_key, expected):
    config = {
        'a.b.foo': 1,
        'a.b.bar': 2,
        'c.d.foo': 3,
        'e.f.o': 4,
        'x.y.z': 5,
        'w.x.y.z': 6,
    }
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            fuzzy_lookup(config, fuzzy_key)
    else:
        assert fuzzy_lookup(config, fuzzy_key) == expected


def test_user_override(config):
    apply_user_overrides(
        config, [('biz', '12'), ('e.f', 'range(4)'), ('g.h.i', 'zip("abc", range(3))')]
    )
    assert config['foo.bar.biz'] == 12.0
    assert config['d.e.f'] == [0, 1, 2, 3]
    assert config['g.h.i'] == {'a': 0, 'b': 1, 'c': 2}


def test_user_override_type_mismatch(config):
    with pytest.raises(ConfigError):
        apply_user_overrides(config, [('d.e.f', 'os.system("clear")')])


def test_user_override_invalid_value(config):
    with pytest.raises(ConfigError):
        apply_user_overrides(config, [('baz', '1')])


def test_user_override_invalid_key(config):
    with pytest.raises(ConfigError):
        apply_user_overrides(config, [('not.a.key', '1')])


def test_user_override_int(config):
    apply_user_overrides(config, [('bar.baz', '18')])
    assert config['foo.bar.baz'] == 18


def test_user_override_int_invalid(config):
    with pytest.raises(ConfigError):
        apply_user_overrides(config, [('bar.baz', 'eighteen')])


def test_user_override_bool(config):
    apply_user_overrides(config, [('def.baz', '1')])
    assert config['abc.def.baz'] is True
    apply_user_overrides(config, [('def.baz', 'True')])
    assert config['abc.def.baz'] is True
    apply_user_overrides(config, [('def.baz', 'False')])
    assert config['abc.def.baz'] is False


def test_user_override_str(config):
    apply_user_overrides(config, [('a.b.c', 'just a string')])
    assert config['a.b.c'] == 'just a string'


def test_user_override_str_int(config):
    apply_user_overrides(config, [('a.b.c', '123')])
    assert config['a.b.c'] == '123'


def test_user_config(config):
    user_config = {
        'foo.bar.baz': 99,
        'g.h.i': {'c': 1, 'd': 2},
    }
    apply_user_config(config, user_config)
    assert config['foo.bar.baz'] == 99
    assert config['g.h.i'] == {'c': 1, 'd': 2}


def test_user_config_bad_key(config):
    user_config = {'a.bad.key': 1}
    with pytest.raises(ConfigError):
        apply_user_config(config, user_config)


def test_user_config_bad_value(config):
    user_config = {'foo.bar.baz': 'not an int'}
    with pytest.raises(ConfigError):
        apply_user_config(config, user_config)


@pytest.mark.skipif(
    hasattr(sys, 'pypy_version_info'), reason="PyPy's eval() mishandles locals dict"
)
def test_safe_eval_str_builtin_alias():
    assert _safe_eval('oct', str) == 'oct'
    assert _safe_eval('oct') is oct
    with pytest.raises(ConfigError):
        _safe_eval('oct', eval_locals={})
    assert _safe_eval('oct', str, {}) == 'oct'


def test_safe_eval_dict():
    with pytest.raises(ConfigError):
        _safe_eval('oct', coerce_type=dict)


@pytest.mark.parametrize(
    'user_keys, user_exprs, expected',
    [
        ('foo', '1,2,3', (['a.b.foo'], [[1], [2], [3]])),
        ('bar', '1.2, 3, 4.5', (['a.b.bar'], [[1.2], [3.0], [4.5]])),
        ('b.baz', '"abc"', (['a.b.baz'], [['a'], ['b'], ['c']])),
        ('b.baz', '"abc","def"', (['a.b.baz'], [['abc'], ['def']])),
        ('d.baz', '1, "y", 0', (['c.d.baz'], [[True], [True], [False]])),
        ('foo,bar', '(1,1),(2,2)', (['a.b.foo', 'a.b.bar'], [[1, 1.0], [2, 2.0]])),
    ],
)
def test_parse_user_factor(user_keys, user_exprs, expected):
    config = {
        'a.b.foo': 1,
        'a.b.bar': 2.0,
        'a.b.baz': 'three',
        'c.d.baz': True,
    }

    factor = parse_user_factor(config, user_keys, user_exprs)
    assert expected == factor
    assert all(
        isinstance(value, type(expected_value))
        for value, expected_value in zip(factor[1], expected[1])
    )


def test_parse_user_factors(config):
    config = {
        'a.b.foo': 1,
        'a.b.bar': 2.0,
        'a.b.baz': 'three',
        'c.d.baz': True,
    }

    user_factors = [['foo', '1,2,3'], ['bar', '2.0, 4.0']]

    factors = parse_user_factors(config, user_factors)

    assert factors[0] == (['a.b.foo'], [[1], [2], [3]])
    assert factors[1] == (['a.b.bar'], [[2.0], [4.0]])


@pytest.mark.parametrize(
    'user_keys, user_exprs, err_str',
    [
        ('baz', 'True, False', 'ambiguous'),
        ('foo', '"one", "two"', 'coerce'),
        ('foo', '1', 'sequence'),
    ],
)
def test_parse_user_factor_invalid(user_keys, user_exprs, err_str):
    config = {
        'a.b.foo': 1,
        'a.b.bar': 2.0,
        'a.b.baz': 'three',
        'c.d.baz': True,
    }
    with pytest.raises(ConfigError) as e:
        parse_user_factor(config, user_keys, user_exprs)
    print(e)
    assert err_str in str(e)


def test_factorial_config():
    factors = [
        (['k0', 'k1'], [[0, 1], [2, 3]]),
        (['k2'], [[4], [5], [6]]),
    ]

    expected = [
        {'k0': 0, 'k1': 1, 'k2': 4},
        {'k0': 0, 'k1': 1, 'k2': 5},
        {'k0': 0, 'k1': 1, 'k2': 6},
        {'k0': 2, 'k1': 3, 'k2': 4},
        {'k0': 2, 'k1': 3, 'k2': 5},
        {'k0': 2, 'k1': 3, 'k2': 6},
    ]

    assert list(factorial_config({}, factors)) == expected


def test_factorial_config_special():
    factors = [
        (['k0', 'k1'], [[0, 1], [2, 3]]),
        (['k2'], [[4], [5], [6]]),
    ]

    expected = [
        {'k0': 0, 'k1': 1, 'k2': 4, 'special': [('k0', 0), ('k1', 1), ('k2', 4)]},
        {'k0': 0, 'k1': 1, 'k2': 5, 'special': [('k0', 0), ('k1', 1), ('k2', 5)]},
        {'k0': 0, 'k1': 1, 'k2': 6, 'special': [('k0', 0), ('k1', 1), ('k2', 6)]},
        {'k0': 2, 'k1': 3, 'k2': 4, 'special': [('k0', 2), ('k1', 3), ('k2', 4)]},
        {'k0': 2, 'k1': 3, 'k2': 5, 'special': [('k0', 2), ('k1', 3), ('k2', 5)]},
        {'k0': 2, 'k1': 3, 'k2': 6, 'special': [('k0', 2), ('k1', 3), ('k2', 6)]},
    ]

    fc = factorial_config({}, factors, 'special')
    assert list(fc) == expected



import pytest
import simpy

from desmod.pool import Pool, PriorityPool
from desmod.probe import attach
from desmod.queue import Queue


@pytest.fixture
def env():
    return simpy.Environment()


def test_attach_bad_type(env):
    values = []
    with pytest.raises(TypeError):
        attach('scope', 'a string', [values.append])


def test_attach_method():
    values = []

    class C:
        def __init__(self):
            self.x = 0

        def doit(self):
            self.x += 1
            return self.x

    c = C()
    attach('scope', c.doit, [values.append])
    c.doit()
    c.doit()
    c.doit()
    assert values == [1, 2, 3]


def test_attach_container(env):
    values = []
    container = simpy.Container(env)
    attach('scope', container, [values.append])

    def proc():
        yield container.put(2)
        yield container.get(1)

    env.process(proc())
    env.run()
    assert values == [2, 1]


def test_attach_store(env):
    values = []
    store = simpy.Store(env)
    attach('scope', store, [values.append])

    def proc():
        yield store.put('item0')
        yield store.put('item1')
        yield store.put('item2')
        item = yield store.get()
        assert item == 'item0'

    env.process(proc())
    env.run()
    assert values == [1, 2, 3, 2]


def test_attach_resource_users(env):
    values = []
    resource = simpy.Resource(env, capacity=3)
    attach('scope', resource, [values.append])

    def proc():
        with resource.request() as req:
            yield req
            with resource.request() as req:
                yield req
            with resource.request() as req:
                yield req

    env.process(proc())
    env.run()
    assert values == [1, 2, 1, 2, 1, 0]


def test_attach_resource_queue(env):
    values = []
    resource = simpy.Resource(env)
    attach('scope', resource, [values.append], trace_queue=True)

    def proc(t):
        with resource.request() as req:
            yield req
            yield env.timeout(t)

    env.process(proc(1))
    env.process(proc(2))
    env.process(proc(3))
    env.run()
    assert values == [0, 1, 2, 1, 0]


def test_attach_queue_size(env):
    values = []
    queue = Queue(env)
    attach('scope', queue, [values.append])

    def proc():
        yield queue.put('item0')
        yield queue.put('item1')
        yield queue.put('item2')
        item = yield queue.get()
        assert item == 'item0'

    env.process(proc())
    env.run()
    assert values == [1, 2, 3, 2]


def test_attach_queue_remaining(env):
    values = []
    queue = Queue(env, capacity=10)

    attach('scope', queue, [values.append], trace_remaining=True)

    def proc():
        yield queue.put('item0')
        yield queue.put('item1')
        yield queue.put('item2')
        item = yield queue.get()
        assert item == 'item0'

    env.process(proc())
    env.run()
    assert values == [9, 8, 7, 8]


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_attach_pool_level(env, PoolClass):
    values = []
    pool = PoolClass(env)
    attach('scope', pool, [values.append])

    def proc():
        yield pool.put(1)
        yield pool.put(1)
        yield pool.put(1)
        item = yield pool.get(1)
        assert item == 1

    env.process(proc())
    env.run()
    assert values == [1, 2, 3, 2]


@pytest.mark.parametrize('PoolClass', [Pool, PriorityPool])
def test_attach_pool_remaining(env, PoolClass):
    values = []
    pool = PoolClass(env, capacity=10)

    attach('scope', pool, [values.append], trace_remaining=True)

    def proc():
        yield pool.put(1)
        yield pool.put(1)
        yield pool.put(1)
        item = yield pool.get(3)
        assert item == 3

    env.process(proc())
    env.run()
    assert values == [9, 8, 7, 10]



import os

import pytest

from desmod.component import Component
from desmod.dot import component_to_dot, generate_dot
from desmod.simulation import SimEnvironment

pytestmark = pytest.mark.usefixtures('cleandir')


@pytest.fixture
def cleandir(tmpdir):
    origin = os.getcwd()
    tmpdir.chdir()
    yield None
    os.chdir(origin)


@pytest.fixture
def top():
    top = Top(parent=None, env=SimEnvironment(config={}))
    top.elaborate()
    return top


class Top(Component):
    base_name = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.a = A(self)
        self.bs = [B(self, index=i) for i in range(5)]

    def connect_children(self):
        for b in self.bs:
            self.connect(b, 'a')


class A(Component):
    base_name = 'a'


class B(Component):
    base_name = 'b'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_connections('a')
        self.add_process(self.my_proc)

    def my_proc(self):
        yield self.env.timeout(1)  # pragma: no coverage


def test_hierarchy_only(top):
    dot = component_to_dot(top, show_connections=False, show_processes=False)
    assert '"a"' in dot
    assert '"b0"' in dot


def test_connections_only(top):
    dot = component_to_dot(top, show_hierarchy=False, show_processes=False)
    assert '"b0" -> "a"' in dot


def test_processes_only(top):
    dot = component_to_dot(top, show_hierarchy=False, show_connections=False)
    assert 'my_proc' in dot


def test_all(top):
    dot = component_to_dot(top, colorscheme='blues9')
    assert 'my_proc' in dot
    assert '"a"' in dot
    assert '"b0"' in dot
    assert '"b0" -> "a"' in dot


@pytest.mark.parametrize(
    'key',
    [
        'sim.dot.enable',
        'sim.dot.colorscheme',
        'sim.dot.all.file',
        'sim.dot.hier.file',
        'sim.dot.conn.file',
    ],
)
def test_generate_dot(top, key):
    assert key not in top.env.config
    generate_dot(top)
    assert key in top.env.config
    files = os.listdir(os.curdir)
    for key in top.env.config:
        if key.startswith('sim.dot.') and key.endswith('.file'):
            assert top.env.config[key] not in files


@pytest.mark.parametrize(
    'key', ['sim.dot.all.file', 'sim.dot.hier.file', 'sim.dot.conn.file']
)
def test_generate_dot_file_enables(top, key):
    top.env.config['sim.dot.enable'] = True
    top.env.config[key] = ''
    generate_dot(top)
    assert all(name.endswith('.dot') for name in os.listdir(os.curdir))



from typing import TYPE_CHECKING, Any, Callable, Generator, List, Optional, Union
import os
import re
import sqlite3
import sys
import traceback

from vcd import VCDWriter
import simpy

from .pool import Pool
from .probe import ProbeCallback, ProbeTarget
from .probe import attach as probe_attach
from .queue import Queue
from .timescale import parse_time, scale_time
from .util import partial_format

if TYPE_CHECKING:
    from .simulation import SimEnvironment

TraceCallback = Callable[..., None]


class Tracer:

    name: str = ''

    def __init__(self, env: 'SimEnvironment'):
        self.env = env
        cfg_scope = f'sim.{self.name}'
        self.enabled: bool = env.config.setdefault(f'{cfg_scope}.enable', False)
        self.persist: bool = env.config.setdefault(f'{cfg_scope}.persist', True)
        if self.enabled:
            self.open()
            include_pat: List[str] = env.config.setdefault(
                f'{cfg_scope}.include_pat', ['.*']
            )
            exclude_pat: List[str] = env.config.setdefault(
                f'{cfg_scope}.exclude_pat', []
            )
            self._include_re = [re.compile(pat) for pat in include_pat]
            self._exclude_re = [re.compile(pat) for pat in exclude_pat]

    def is_scope_enabled(self, scope: str) -> bool:
        return (
            self.enabled
            and any(r.match(scope) for r in self._include_re)
            and not any(r.match(scope) for r in self._exclude_re)
        )

    def open(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    def close(self) -> None:
        if self.enabled:
            self._close()

    def _close(self) -> None:
        raise NotImplementedError()  # pragma: no cover

    def remove_files(self) -> None:
        raise NotImplementedError()

    def flush(self) -> None:
        pass

    def activate_probe(
        self, scope: str, target: ProbeTarget, **hints: Any
    ) -> Optional[ProbeCallback]:
        raise NotImplementedError()  # pragma: no cover

    def activate_trace(self, scope: str, **hints) -> Optional[TraceCallback]:
        raise NotImplementedError()  # pragma: no cover

    def trace_exception(self) -> None:
        pass


class LogTracer(Tracer):

    name = 'log'
    default_format = '{level:7} {ts:.3f} {ts_unit}: {scope}:'

    levels = {
        'ERROR': 1,
        'WARNING': 2,
        'INFO': 3,
        'PROBE': 4,
        'DEBUG': 5,
    }

    def open(self) -> None:
        self.filename: str = self.env.config.setdefault('sim.log.file', 'sim.log')
        buffering: int = self.env.config.setdefault('sim.log.buffering', -1)
        level: str = self.env.config.setdefault('sim.log.level', 'INFO')
        self.max_level = self.levels[level]
        self.format_str: str = self.env.config.setdefault(
            'sim.log.format', self.default_format
        )
        ts_n, ts_unit = self.env.timescale
        if ts_n == 1:
            self.ts_unit = ts_unit
        else:
            self.ts_unit = f'({ts_n}{ts_unit})'

        if self.filename:
            self.file = open(self.filename, 'w', buffering)
            self.should_close = True
        else:
            self.file = sys.stderr
            self.should_close = False

    def flush(self) -> None:
        self.file.flush()

    def _close(self) -> None:
        if self.should_close:
            self.file.close()

    def remove_files(self) -> None:
        if os.path.isfile(self.filename):
            os.remove(self.filename)

    def is_scope_enabled(self, scope: str, level: Optional[str] = None) -> bool:
        return (level is None or self.levels[level] <= self.max_level) and super(
            LogTracer, self
        ).is_scope_enabled(scope)

    def activate_probe(
        self, scope: str, target: ProbeTarget, **hints: Any
    ) -> Optional[ProbeCallback]:
        level: str = hints.get('level', 'PROBE')
        if not self.is_scope_enabled(scope, level):
            return None
        format_str = partial_format(
            self.format_str, level=level, ts_unit=self.ts_unit, scope=scope
        )

        def probe_callback(value: object) -> None:
            print(format_str.format(ts=self.env.now), value, file=self.file)

        return probe_callback

    def activate_trace(self, scope: str, **hints) -> Optional[TraceCallback]:
        level: str = hints.get('level', 'DEBUG')
        if not self.is_scope_enabled(scope, level):
            return None
        format_str = partial_format(
            self.format_str, level=level, ts_unit=self.ts_unit, scope=scope
        )

        def trace_callback(*value) -> None:
            print(format_str.format(ts=self.env.now), *value, file=self.file)

        return trace_callback

    def trace_exception(self) -> None:
        tb_lines = traceback.format_exception(*sys.exc_info())
        print(
            self.format_str.format(
                level='ERROR', ts=self.env.now, ts_unit=self.ts_unit, scope='Exception'
            ),
            tb_lines[-1],
            '\n',
            *tb_lines,
            file=self.file,
        )


class VCDTracer(Tracer):

    name = 'vcd'

    def open(self) -> None:
        dump_filename: str = self.env.config.setdefault('sim.vcd.dump_file', 'sim.vcd')
        if 'sim.vcd.timescale' in self.env.config:
            vcd_ts_str: str = self.env.config.setdefault(
                'sim.vcd.timescale', self.env.config['sim.timescale']
            )
            mag, unit = parse_time(vcd_ts_str)
        else:
            mag, unit = self.env.timescale
        mag_int = int(mag)
        if mag_int != mag:
            raise ValueError(f'sim.timescale magnitude must be an integer, got {mag}')
        vcd_timescale = mag_int, unit
        self.scale_factor = scale_time(self.env.timescale, vcd_timescale)
        check_values: bool = self.env.config.setdefault('sim.vcd.check_values', True)
        self.dump_file = open(dump_filename, 'w')
        self.vcd = VCDWriter(
            self.dump_file, timescale=vcd_timescale, check_values=check_values
        )
        self.save_filename: str = self.env.config.setdefault(
            'sim.gtkw.file', 'sim.gtkw'
        )
        if self.env.config.setdefault('sim.gtkw.live'):
            from vcd.gtkw import spawn_gtkwave_interactive

            quiet: bool = self.env.config.setdefault('sim.gtkw.quiet', True)
            spawn_gtkwave_interactive(dump_filename, self.save_filename, quiet=quiet)

        start_time: str = self.env.config.setdefault('sim.vcd.start_time', '')
        stop_time: str = self.env.config.setdefault('sim.vcd.stop_time', '')
        t_start = (
            scale_time(parse_time(start_time), self.env.timescale)
            if start_time
            else None
        )
        t_stop = (
            scale_time(parse_time(stop_time), self.env.timescale) if stop_time else None
        )
        self.env.process(self._start_stop(t_start, t_stop))

    def vcd_now(self) -> float:
        return self.env.now * self.scale_factor

    def flush(self) -> None:
        self.dump_file.flush()

    def _close(self) -> None:
        self.vcd.close(self.vcd_now())
        self.dump_file.close()

    def remove_files(self) -> None:
        if os.path.isfile(self.dump_file.name):
            os.remove(self.dump_file.name)
        if os.path.isfile(self.save_filename):
            os.remove(self.save_filename)

    def activate_probe(
        self, scope: str, target: ProbeTarget, **hints: Any,
    ) -> Optional[ProbeCallback]:
        assert self.enabled
        var_type: Optional[str] = hints.get('var_type')
        if var_type is None:
            if isinstance(target, (simpy.Container, Pool)):
                if isinstance(target.level, float):
                    var_type = 'real'
                else:
                    var_type = 'integer'
            elif isinstance(target, (simpy.Resource, simpy.Store, Queue)):
                var_type = 'integer'
            else:
                raise ValueError(f'Could not infer VCD var_type for {scope}')

        kwargs = {k: hints[k] for k in ['size', 'init', 'ident'] if k in hints}

        if 'init' not in kwargs:
            if isinstance(target, (simpy.Container, Pool)):
                kwargs['init'] = target.level
            elif isinstance(target, simpy.Resource):
                kwargs['init'] = len(target.users) if target.users else 'z'
            elif isinstance(target, (simpy.Store, Queue)):
                kwargs['init'] = len(target.items)

        parent_scope, name = scope.rsplit('.', 1)
        var = self.vcd.register_var(parent_scope, name, var_type, **kwargs)

        def probe_callback(value: Any) -> None:
            self.vcd.change(var, self.vcd_now(), value)

        return probe_callback

    def activate_trace(self, scope: str, **hints) -> Optional[TraceCallback]:
        assert self.enabled
        var_type = hints['var_type']
        kwargs = {k: hints[k] for k in ['size', 'init', 'ident'] if k in hints}

        parent_scope, name = scope.rsplit('.', 1)
        var = self.vcd.register_var(parent_scope, name, var_type, **kwargs)

        if isinstance(var.size, tuple):

            def trace_callback(*value) -> None:
                self.vcd.change(var, self.vcd_now(), value)

        else:

            def trace_callback(*value) -> None:
                self.vcd.change(var, self.vcd_now(), value[0])

        return trace_callback

    def _start_stop(
        self, t_start: Optional[Union[int, float]], t_stop: Optional[Union[int, float]],
    ) -> Generator[simpy.Timeout, None, None]:
        # Wait for simulation to start to ensure all variable registration is
        # complete before doing and dump_on()/dump_off() calls.
        yield self.env.timeout(0)

        if t_start is None and t_stop is None:
            # |vvvvvvvvvvvvvv|
            pass
        elif t_start is None:
            assert t_stop is not None
            # |vvvvvv--------|
            yield self.env.timeout(t_stop)
            self.vcd.dump_off(self.vcd_now())
        elif t_stop is None:
            # |--------vvvvvv|
            self.vcd.dump_off(self.vcd_now())
            yield self.env.timeout(t_start)
            self.vcd.dump_on(self.vcd_now())
        elif t_start <= t_stop:
            # |---vvvvvv-----|
            self.vcd.dump_off(self.vcd_now())
            yield self.env.timeout(t_start)
            self.vcd.dump_on(self.vcd_now())
            yield self.env.timeout(t_stop - t_start)
            self.vcd.dump_off(self.vcd_now())
        else:
            # |vvv-------vvvv|
            yield self.env.timeout(t_stop)
            self.vcd.dump_off(self.vcd_now())
            yield self.env.timeout(t_start - t_stop)
            self.vcd.dump_on(self.vcd_now())


class SQLiteTracer(Tracer):

    name = 'db'

    def open(self) -> None:
        self.filename: str = self.env.config.setdefault('sim.db.file', 'sim.sqlite')
        self.trace_table: str = self.env.config.setdefault(
            'sim.db.trace_table', 'trace'
        )
        self.remove_files()
        self.db = sqlite3.connect(self.filename)
        self._is_trace_table_created = False

    def _create_trace_table(self) -> None:
        if not self._is_trace_table_created:
            self.db.execute(
                f'CREATE TABLE {self.trace_table} ('
                f'timestamp FLOAT, '
                f'scope TEXT, '
                f'value)'
            )
            self._is_trace_table_created = True

    def flush(self) -> None:
        self.db.commit()

    def _close(self) -> None:
        self.db.commit()
        self.db.close()

    def remove_files(self) -> None:
        if self.filename != ':memory:':
            for filename in [self.filename, f'{self.filename}-journal']:
                if os.path.exists(filename):
                    os.remove(filename)

    def activate_probe(
        self, scope: str, target: ProbeTarget, **hints: Any
    ) -> Optional[ProbeCallback]:
        return self.activate_trace(scope, **hints)

    def activate_trace(self, scope: str, **hints) -> Optional[TraceCallback]:
        assert self.enabled
        self._create_trace_table()
        insert_sql = (
            f'INSERT INTO {self.trace_table} (timestamp, scope, value) VALUES (?, ?, ?)'
        )

        def trace_callback(value) -> None:
            self.db.execute(insert_sql, (self.env.now, scope, value))

        return trace_callback


class TraceManager:
    def __init__(self, env: 'SimEnvironment') -> None:
        self.tracers: List[Tracer] = []
        try:
            self.log_tracer = LogTracer(env)
            self.tracers.append(self.log_tracer)
            self.vcd_tracer = VCDTracer(env)
            self.tracers.append(self.vcd_tracer)
            self.sqlite_tracer = SQLiteTracer(env)
            self.tracers.append(self.sqlite_tracer)
        except BaseException:
            self.close()
            raise

    def flush(self) -> None:
        """Flush all managed tracers instances.

        The effect of flushing is tracer-dependent.

        """
        for tracer in self.tracers:
            if tracer.enabled:
                tracer.flush()

    def close(self) -> None:
        for tracer in self.tracers:
            tracer.close()
            if tracer.enabled and not tracer.persist:
                tracer.remove_files()

    def auto_probe(self, scope: str, target: ProbeTarget, **hints: Any) -> None:
        callbacks: List[ProbeCallback] = []
        for tracer in self.tracers:
            if tracer.name in hints and tracer.is_scope_enabled(scope):
                callback = tracer.activate_probe(scope, target, **hints[tracer.name])
                if callback:
                    callbacks.append(callback)
        if callbacks:
            probe_attach(scope, target, callbacks, **hints)

    def get_trace_function(self, scope: str, **hints) -> Callable[..., None]:
        callbacks = []
        for tracer in self.tracers:
            if tracer.name in hints and tracer.is_scope_enabled(scope):
                callback = tracer.activate_trace(scope, **hints[tracer.name])
                if callback:
                    callbacks.append(callback)

        def trace_function(*value) -> None:
            for callback in callbacks:
                callback(*value)

        return trace_function

    def trace_exception(self) -> None:
        for tracer in self.tracers:
            if tracer.enabled:
                tracer.trace_exception()



"""Component is the building block for desmod models.

Hierarchy
---------

A desmod model consists of a directed acyclical graph (DAG) of
:class:`Component` subclasses. Each Component is composed of zero or more child
Components. A single top-level Component class is passed to the
:func:`~desmod.simulation.simulate()` function to initiate simulation.

The :class:`Component` hierarchy does not define the behavior of a model, but
instead exists as a tool to build large models out of composable and
encapsulated pieces.

Connections
-----------

Components connect to other components via connection objects. Each component
is responsible for declaring the names of external connections as well as make
connections for its child components. The final network of inter-component
connections is neither directed (a connection object may enable two-way
communication), acyclic (groups of components may form cyclical connections),
nor constrained to match the component hierarchy.

Ultimately, a connection between two components means that each component
instance has a [pythonic] reference to the connection object.

In the spirit of Python, the types connection objects are flexible and dynamic.
A connection object may be of any type--it is up to the connected components to
cooperatively decide how to use the connection object for communication. That
said, some object types are more useful than others for connections. Some
useful connection object types include:

 * :class:`desmod.queue.Queue`
 * :class:`simpy.resources.resource.Resource`

Processes
---------

A component may have zero or more simulation processes
(:class:`simpy.events.Process`). It is these processes that give a model its
simulation-time behavior. The process methods declared by components are
started at simulation time. These "standing" processes may dynamically launch
addtional processes using `self.env.process()`.

Use Cases
---------

Given the flexibility components to have zero or more children, zero or more
processes, and zero or more connections, it can be helpful to give names to
the various roles components may play in a model.

 * Structural Component -- a component with child components, but no processes
 * Behavioral Component -- a component with processes, but no child components
 * Hybrid Component -- a component with child components and processes
 * State Component -- a component with neither children or processes

It is typical for the top-level component in a model to be purely structural,
while behavioral components are leaves in the model DAG.

A component with neither children or processes may still be useful. Such a
component could, for example, be used as a connection object.

"""

from typing import Any, Callable, Dict, Generator, List, Optional, Set, Tuple

import simpy

from .simulation import ResultDict, SimEnvironment

ProcessGenerator = Callable[..., Generator[simpy.Event, Any, None]]


class ConnectError(Exception):
    pass


class Component:
    """Building block for composing models.

    This class is meant to be subclassed. Component subclasses must declare
    their children, connections, and processes.

    :param Component parent: Parent component or None for top-level Component.
    :param SimEnvironment env: SimPy simulation environment.
    :param str name: Optional name of Component instance.
    :param int index:
        Optional index of Component. This is used when multiple sibling
        components of the same type are instantiated as an array/list.

    """

    #: Short/friendly name used in the scope (class attribute).
    base_name: str = ''

    def __init__(
        self,
        parent: Optional['Component'],
        env: Optional[SimEnvironment] = None,
        name: Optional[str] = None,
        index: Optional[int] = None,
    ) -> None:
        #: The simulation environment; a :class:`SimEnvironment` instance.
        self.env: SimEnvironment
        if env is not None:
            self.env = env
        elif parent is not None:
            self.env = parent.env
        else:
            raise AssertionError('either parent or env must be non-None')

        #: The component name (str).
        self.name = (self.base_name if name is None else name) + (
            '' if index is None else str(index)
        )

        #: Index of Component instance within group of sibling instances.
        #: Will be None for un-grouped Components.
        self.index = index

        #: String indicating the full scope of Component instance in the
        #: Component DAG.
        self.scope: str
        if parent is None or not parent.scope:
            self.scope = self.name
        else:
            self.scope = f'{parent.scope}.{self.name}'

        if parent:
            parent._children.append(self)

        self._children: List['Component'] = []
        self._processes: List[
            Tuple[ProcessGenerator, Tuple[Any, ...], Dict[str, Any]]
        ] = []
        self._connections: List[Any] = []
        self._not_connected: Set[str] = set()

        #: Log an error message.
        self.error: Callable[..., None] = self.env.tracemgr.get_trace_function(
            self.scope, log={'level': 'ERROR'}
        )
        #: Log a warning message.
        self.warn: Callable[..., None] = self.env.tracemgr.get_trace_function(
            self.scope, log={'level': 'WARNING'}
        )
        #: Log an informative message.
        self.info: Callable[..., None] = self.env.tracemgr.get_trace_function(
            self.scope, log={'level': 'INFO'}
        )
        #: Log a debug message.
        self.debug: Callable[..., None] = self.env.tracemgr.get_trace_function(
            self.scope, log={'level': 'DEBUG'}
        )

    def add_process(self, g: ProcessGenerator, *args: Any, **kwargs: Any) -> None:
        """Add a process method to be run at simulation-time.

        Subclasses should call this in `__init__()` to declare the process
        methods to be started at simulation-time.

        :param function process_func:
            Typically a bound method of the Component subclass.
        :param args: arguments to pass to `process_func`.
        :param kwargs: keyword arguments to pass to `process_func`.

        """
        self._processes.append((g, args, kwargs))

    def add_processes(self, *generators: ProcessGenerator) -> None:
        """Declare multiple processes at once.

        This is a convenience wrapper for :meth:`add_process()` that may be
        used to quickly declare a list of process methods that do not require
        any arguments.

        :param process_funcs: argument-less process functions (methods).

        """
        for g in generators:
            self.add_process(g)

    def add_connections(self, *connection_names: str) -> None:
        """Declare names of externally-provided connection objects.

        The named connections must be connected (assigned) by an ancestor at
        elaboration time.

        """
        self._not_connected.update(connection_names)

    def connect(
        self,
        dst: 'Component',
        dst_connection: Any,
        src: Optional['Component'] = None,
        src_connection: Optional[Any] = None,
        conn_obj: Optional[Any] = None,
    ) -> None:
        """Assign connection object from source to destination component.

        At elaboration-time, Components must call `connect()` to make the
        connections declared by descendant (child, grandchild, etc.)
        components.

        .. Note::

            :meth:`connect()` is nominally called from
            :meth:`connect_children()`.

        :param Component dst:
            Destination component being assigned the connection object.
        :param str dst_connection:
            Destination's name for the connection object.
        :param Component src:
            Source component providing the connection object. If omitted, the
            source component is assumed to be `self`.
        :param str src_connection:
            Source's name for the connection object. If omitted,
            `dst_connection` is used.
        :param conn_obj:
            The connection object to be assigned to the destination component.
            This parameter may typically be omitted in which case the
            connection object is resolved using `src` and `src_connection`.

        """
        if src is None:
            src = self
        if src_connection is None:
            src_connection = dst_connection
        if conn_obj is None:
            if hasattr(src, src_connection):
                conn_obj = getattr(src, src_connection)
            else:
                raise ConnectError(
                    f'src "{src.scope}" (class {type(src).__name__}) does not have attr'
                    f'"{src_connection}"'
                )
        if dst_connection in dst._not_connected:
            setattr(dst, dst_connection, conn_obj)
            dst._not_connected.remove(dst_connection)
            dst._connections.append((dst_connection, src, src_connection, conn_obj))
        else:
            raise ConnectError(
                f'dst "{dst.scope}" (class {type(dst).__name__}) does not declare'
                f'connection "{dst_connection}"'
            )

    def connect_children(self) -> None:
        """Make connections for descendant components.

        This method must be overridden in Component subclasses that need to
        make any connections on behalf of its descendant components.
        Connections are made using :meth:`connect()`.

        """
        if any(child._not_connected for child in self._children):
            raise ConnectError(
                '{0} has unconnected children; implement '
                '{0}.connect_children()'.format(type(self).__name__)
            )

    def auto_probe(self, name: str, target: Any = None, **hints: Any) -> None:
        if target is None:
            target = getattr(self, name)
        target_scope = '.'.join([self.scope, name])
        self.env.tracemgr.auto_probe(target_scope, target, **hints)

    def get_trace_function(self, name: str, **hints: Any) -> Callable[..., None]:
        target_scope = '.'.join([self.scope, name])
        return self.env.tracemgr.get_trace_function(target_scope, **hints)

    @classmethod
    def pre_init(cls, env: SimEnvironment) -> None:
        """Override-able class method called prior to model initialization.

        Component subclasses may override this classmethod to gain access
        to the simulation environment (`env`) prior to :meth:`__init__()` being
        called.

        """
        pass

    def elaborate(self) -> None:
        """Recursively elaborate the model.

        The elaboration phase prepares the model for simulation. Descendant
        connections are made and components' processes are started at
        elaboration-time.

        """
        self.connect_children()
        for child in self._children:
            if child._not_connected:
                raise ConnectError(
                    f'{child.scope}.{child._not_connected.pop()} not connected'
                )
            child.elaborate()
        for proc, args, kwargs in self._processes:
            self.env.process(proc(*args, **kwargs))
        self.elab_hook()

    def elab_hook(self) -> None:
        """Hook called after elaboration and before simulation phase.

        Component subclasses may override :meth:`elab_hook()` to inject
        behavior after elaboration, but prior to simulation.

        """
        pass

    def post_simulate(self) -> None:
        """Recursively run post-simulation hooks."""
        for child in self._children:
            child.post_simulate()
        self.post_sim_hook()

    def post_sim_hook(self) -> None:
        """Hook called after simulation completes.

        Component subclasses may override `post_sim_hook()` to inject behavior
        after the simulation completes successfully. Note that
        `post_sim_hook()` will not be called if the simulation terminates with
        an unhandled exception.

        """
        pass

    def get_result(self, result: ResultDict) -> None:
        """Recursively compose simulation result dict.

        Upon successful completion of the simulation phase, each component in
        the model has the opportunity to add-to or modify the `result` dict via
        its :meth:`get_result_hook` method.

        The fully composed `result` dict is returned by :func:`simulate`.

        :param dict result: Result dictionary to be modified.

        """
        for child in self._children:
            child.get_result(result)
        self.get_result_hook(result)

    def get_result_hook(self, result: ResultDict) -> None:
        """Hook called after result is composed by descendant components."""
        pass



from functools import wraps
from types import MethodType
from typing import Any, Callable, Iterable, Union

import simpy

from desmod.pool import Pool
from desmod.queue import ItemType, Queue

ProbeCallback = Callable[[Any], None]
ProbeCallbacks = Iterable[ProbeCallback]
ProbeTarget = Union[
    Pool, Queue[ItemType], simpy.Resource, simpy.Store, simpy.Container, MethodType
]


def attach(
    scope: str, target: ProbeTarget, callbacks: ProbeCallbacks, **hints: Any
) -> None:
    if isinstance(target, MethodType):
        _attach_method(target, callbacks)
    elif isinstance(target, simpy.Container):
        _attach_container_level(target, callbacks)
    elif isinstance(target, simpy.Store):
        _attach_store_items(target, callbacks)
    elif isinstance(target, simpy.Resource):
        if hints.get('trace_queue'):
            _attach_resource_queue(target, callbacks)
        else:
            _attach_resource_users(target, callbacks)
    elif isinstance(target, Queue):
        if hints.get('trace_remaining', False):
            _attach_queue_remaining(target, callbacks)
        else:
            _attach_queue_size(target, callbacks)
    elif isinstance(target, Pool):
        if hints.get('trace_remaining', False):
            _attach_pool_remaining(target, callbacks)
        else:
            _attach_pool_level(target, callbacks)
    else:
        raise TypeError(f'Cannot probe {scope} of type {type(target)}')


def _attach_method(method: MethodType, callbacks: ProbeCallbacks) -> None:
    def make_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            value = func(*args, **kwargs)
            for callback in callbacks:
                callback(value)
            return value

        return wrapper

    setattr(method.__self__, method.__func__.__name__, make_wrapper(method))


def _attach_container_level(
    container: simpy.Container, callbacks: ProbeCallbacks
) -> None:
    def make_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_level = container._level
            ret = func(*args, **kwargs)
            new_level = container._level
            if new_level != old_level:
                for callback in callbacks:
                    callback(new_level)
            return ret

        return wrapper

    container._do_get = make_wrapper(container._do_get)  # type: ignore
    container._do_put = make_wrapper(container._do_put)  # type: ignore


def _attach_store_items(store: simpy.Store, callbacks: ProbeCallbacks) -> None:
    def make_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_items = len(store.items)
            ret = func(*args, **kwargs)
            new_items = len(store.items)
            if new_items != old_items:
                for callback in callbacks:
                    callback(new_items)
            return ret

        return wrapper

    store._do_get = make_wrapper(store._do_get)  # type: ignore
    store._do_put = make_wrapper(store._do_put)  # type: ignore


def _attach_resource_users(resource: simpy.Resource, callbacks: ProbeCallbacks) -> None:
    def make_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_users = len(resource.users)
            ret = func(*args, **kwargs)
            new_users = len(resource.users)
            if new_users != old_users:
                for callback in callbacks:
                    callback(new_users)
            return ret

        return wrapper

    resource._do_get = make_wrapper(resource._do_get)  # type: ignore
    resource._do_put = make_wrapper(resource._do_put)  # type: ignore


def _attach_resource_queue(resource: simpy.Resource, callbacks: ProbeCallbacks) -> None:
    def make_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_queue = len(resource.queue)
            ret = func(*args, **kwargs)
            new_queue = len(resource.queue)
            if new_queue != old_queue:
                for callback in callbacks:
                    callback(new_queue)
            return ret

        return wrapper

    resource.request = make_wrapper(resource.request)  # type: ignore
    resource._trigger_put = make_wrapper(resource._trigger_put)  # type: ignore


def _attach_queue_size(queue: Queue[ItemType], callbacks: ProbeCallbacks) -> None:
    def hook():
        for callback in callbacks:
            callback(queue.size)

    queue._put_hook = queue._get_hook = hook


def _attach_queue_remaining(queue: Queue[ItemType], callbacks: ProbeCallbacks) -> None:
    def hook():
        for callback in callbacks:
            callback(queue.remaining)

    queue._put_hook = queue._get_hook = hook


def _attach_pool_level(pool: Pool, callbacks: ProbeCallbacks) -> None:
    def hook():
        for callback in callbacks:
            callback(pool.level)

    pool._put_hook = pool._get_hook = hook


def _attach_pool_remaining(pool: Pool, callbacks: ProbeCallbacks) -> None:
    def hook():
        for callback in callbacks:
            callback(pool.remaining)

    pool._put_hook = pool._get_hook = hook



"""Generate graphical representation of component hierarchy.

Component hierarchy, connections, and processes can be represented graphically
using the `Graphviz`_ `DOT language`_.

The :func:`component_to_dot()` function produces a DOT language string that can
be rendered into a variety of formats using Graphviz tools.  Because the
component hierarchy, connections, and processes are determined dynamically,
:func:`component_to_dot()` must be called with an instantiated component. A
good way to integrate this capabililty into a model is to call
:func:`component_to_dot()` from a component's
:meth:`desmod.component.Component.elab_hook()` method.

The ``dot`` program from `Graphviz`_ may be used to render the generated DOT
language description of the component hierarchy::

    dot -Tpng -o foo.png foo.dot

For large component hierarchies, the ``osage`` program (also part of Graphviz)
can produce a more compact layout::

    osage -Tpng -o foo.png foo.dot

.. _Graphviz: http://graphviz.org/
.. _DOT language: http://graphviz.org/content/dot-language

"""
from itertools import cycle, groupby
from typing import Dict, Iterator, List, Optional, Sequence

from desmod.component import Component
from desmod.config import ConfigDict

_color_cycle = cycle(
    [
        'dodgerblue4',
        'darkgreen',
        'darkorchid',
        'darkslategray',
        'deeppink4',
        'goldenrod4',
        'firebrick4',
    ]
)


def generate_dot(top: Component, config: Optional[ConfigDict] = None) -> None:
    """Generate dot files based on 'sim.dot' configuration.

    The ``sim.dot.enable`` configuration controls whether any dot file
    generation is performed. The remaining ``sim.dot`` configuration items have
    no effect unless ``sim.dot.enable`` is ``True``.

    The ``sim.dot.colorscheme`` configuration controls the colorscheme used in
    the generated DOT files. See :func:`component_to_dot` for more detail.

    The ``sim.dot.all.file``, ``sim.dot.hier.file``, and ``sim.dot.conn.file``
    configuration items control the names of the generated DOT files. These
    items can also be set to the empty string to disable generating a
    particular file.

    The nominal way to use this function is to call it from the top component's
    :meth:`Component.elab_hook()`. E.g.::

        def elab_hook(self):
            ...
            generate_dot(self)
            ...

    """
    config = top.env.config if config is None else config

    enable: bool = config.setdefault('sim.dot.enable', False)
    colorscheme: str = config.setdefault('sim.dot.colorscheme', '')
    all_filename: str = config.setdefault('sim.dot.all.file', 'all.dot')
    hier_filename: str = config.setdefault('sim.dot.hier.file', 'hier.dot')
    conn_filename: str = config.setdefault('sim.dot.conn.file', 'conn.dot')

    if not enable:
        return

    if all_filename:
        with open(all_filename, 'w') as dot_file:
            dot_file.write(
                component_to_dot(
                    top,
                    show_hierarchy=True,
                    show_connections=True,
                    show_processes=True,
                    colorscheme=colorscheme,
                )
            )

    if hier_filename:
        with open(hier_filename, 'w') as dot_file:
            dot_file.write(
                component_to_dot(
                    top,
                    show_hierarchy=True,
                    show_connections=False,
                    show_processes=False,
                    colorscheme=colorscheme,
                )
            )

    if conn_filename:
        with open(conn_filename, 'w') as dot_file:
            dot_file.write(
                component_to_dot(
                    top,
                    show_hierarchy=False,
                    show_connections=True,
                    show_processes=False,
                    colorscheme=colorscheme,
                )
            )


def component_to_dot(
    top: Component,
    show_hierarchy: bool = True,
    show_connections: bool = True,
    show_processes: bool = True,
    colorscheme: str = '',
) -> str:
    """Produce a dot stream from a component hierarchy.

    The DOT language representation of the component instance hierarchy can
    show the component hierarchy, the inter-component connections, components'
    processes, or any combination thereof.

    .. Note::
        The `top` component hierarchy must be initialized and all connections
        must be made in order for `component_to_dot()` to inspect these graphs.
        The :meth:`desmod.component.Component.elab_hook()` method is a good
        place to call `component_to_dot()` since the model is fully elaborated
        at that point and simulation has not yet started.

    :param Component top: Top-level component (instance).
    :param bool show_hierarchy:
        Should the component hierarchy be shown in the graph.
    :param bool show_connections:
        Should the inter-component connections be shown in the graph.
    :param bool show_processes:
        Should each component's processes be shown in the graph.
    :param str colorscheme:
        One of the `Brewer color schemes`_ supported by graphviz, e.g. "blues8"
        or "set27". Each level of the component hierarchy will use a different
        color from the color scheme. N.B. Brewer color schemes have between 3
        and 12 colors; one should be chosen that has at least as many colors as
        the depth of the component hierarchy.
    :returns str:
        DOT language representation of the component/connection graph(s).

    .. _Brewer color schemes: http://graphviz.org/content/color-names#brewer

    """
    indent = '    '
    lines = ['strict digraph M {']
    lines.extend(
        indent + line
        for line in _comp_hierarchy(
            [top], show_hierarchy, show_connections, show_processes, colorscheme
        )
    )
    if show_connections:
        lines.append('')
        lines.extend(indent + line for line in _comp_connections(top))
    lines.append('}')
    return '\n'.join(lines)


def _comp_hierarchy(
    component_group: Sequence[Component],
    show_hierarchy: bool,
    show_connections: bool,
    show_processes: bool,
    colorscheme: str,
    _level: int = 1,
) -> List[str]:
    component = component_group[0]
    if len(component_group) == 1:
        label_name = _comp_name(component)
    else:
        label_name = (
            f'{_comp_name(component_group[0])}..{_comp_name(component_group[-1])}'
        )

    if component._children and show_hierarchy:
        border_style = 'dotted'
    else:
        border_style = 'rounded'
    if colorscheme:
        style = f'style="{border_style},filled",fillcolor="/{colorscheme}/{_level}"'
    else:
        style = 'style=' + border_style

    node_lines = [f'"{_comp_scope(component)}" [shape=box,{style},label=<']

    label_lines = _comp_label(component, label_name, show_processes)
    if len(label_lines) == 1:
        node_lines[-1] += label_lines[0]
    else:
        node_lines.extend('    ' + line for line in label_lines)
    node_lines[-1] += '>];'

    if not component._children:
        return node_lines
    else:
        if show_hierarchy:
            indent = '    '
            lines = [
                f'subgraph "{_cluster_id(component)}" {{',
                indent + f'{indent}label=<{_cluster_label(component_group)}>',
            ]
            if colorscheme:
                lines.extend(
                    [
                        indent + 'style="filled"',
                        indent + f'fillcolor="/{colorscheme}/{_level}"',
                    ]
                )
        else:
            indent = ''
            lines = []
        if show_connections:
            lines.extend(indent + line for line in node_lines)
        for child_group in _child_type_groups(component):
            lines.extend(
                indent + line
                for line in _comp_hierarchy(
                    child_group,
                    show_hierarchy,
                    show_connections,
                    show_processes,
                    colorscheme,
                    _level + 1,
                )
            )
        if show_hierarchy:
            lines.append('}')
        return lines


def _comp_connections(component: Component) -> List[str]:
    lines = []
    for conn, src, src_conn, conn_obj in component._connections:
        attrs = {}
        if isinstance(conn_obj, Component):
            src = conn_obj
        elif (
            isinstance(conn_obj, list)
            and conn_obj
            and isinstance(conn_obj[0], Component)
        ):
            src = conn_obj[0]
        else:
            attrs['label'] = f'"{conn}"'
            attrs['color'] = attrs['fontcolor'] = next(_color_cycle)

        lines.append(
            '"{dst_id}" -> "{src_id}" [{attrs}];'.format(
                dst_id=_comp_scope(component),
                src_id=_comp_scope(src),
                attrs=_join_attrs(attrs),
            )
        )

    for child_group in _child_type_groups(component):
        lines.extend(_comp_connections(child_group[0]))

    return lines


def _child_type_groups(component: Component) -> Iterator[List[Component]]:
    children = sorted(component._children, key=_comp_name)
    for _, group in groupby(children, lambda child: str(type(child))):
        yield list(group)


def _comp_name(component: Component) -> str:
    return component.name if component.name else type(component).__name__


def _comp_scope(component: Component) -> str:
    return component.scope if component.scope else type(component).__name__


def _cluster_id(component: Component) -> str:
    return 'cluster_' + _comp_scope(component)


def _cluster_label(component_group: Sequence[Component]) -> str:
    if len(component_group) == 1:
        return f'<b>{_comp_name(component_group[0])}</b>'
    else:
        return f'<b>{component_group[0].name}..{component_group[-1].name}</b>'


def _comp_label(
    component: Component, label_name: str, show_processes: bool
) -> List[str]:
    label_lines = [f'<b>{label_name}</b><br align="left"/>']
    if show_processes and component._processes:
        label_lines.append('<br/>')
        proc_funcs = set()
        for proc_func, _, _ in component._processes:
            if proc_func not in proc_funcs:
                proc_funcs.add(proc_func)
                label_lines.append(f'<i>{proc_func.__name__}</i><br align="left"/>')
    return label_lines


def _join_attrs(attrs: Dict[str, str]) -> str:
    return ','.join(f'{k}={v}' for k, v in sorted(attrs.items()))



from contextlib import contextmanager
from datetime import datetime, timedelta
from queue import Queue
from typing import IO, TYPE_CHECKING, Dict, Generator, List, Optional, Set, Tuple, Union
import sys
import timeit

import simpy

from desmod.config import ConfigDict
from desmod.timescale import TimeValue, parse_time, scale_time

try:
    import progressbar
except ImportError:
    progressbar = None
try:
    import colorama
except ImportError:
    colorama = None

if TYPE_CHECKING:
    from desmod.simulation import SimEnvironment

ProgressTuple = Tuple[
    Optional[int],  # simulation index
    Union[int, float],  # now
    Optional[Union[int, float]],  # t_stop
    TimeValue,  # timescale
]


@contextmanager
def standalone_progress_manager(env: 'SimEnvironment') -> Generator[None, None, None]:
    enabled: bool = env.config.setdefault('sim.progress.enable', False)
    max_width: int = env.config.setdefault('sim.progress.max_width')
    period_s = _get_interval_period_s(env.config)

    if enabled:
        if sys.stderr.isatty() and progressbar:
            pbar = _get_standalone_pbar(env, max_width, sys.stderr)
            env.process(_standalone_pbar_process(env, pbar, period_s))
            try:
                yield None
            finally:
                pbar.finish()
        else:
            env.process(_standalone_display_process(env, period_s, sys.stderr))
            try:
                yield None
            finally:
                _print_progress(
                    env.sim_index,
                    env.now,
                    env.now,
                    env.timescale,
                    end='\n',
                    fd=sys.stderr,
                )
    else:
        yield None


def _get_interval_period_s(config: ConfigDict) -> Union[int, float]:
    period_str: str = config.setdefault('sim.progress.update_period', '1 s')
    return scale_time(parse_time(period_str), (1, 's'))


def _standalone_display_process(
    env: 'SimEnvironment', period_s: Union[int, float], fd: IO
) -> Generator[simpy.Timeout, None, None]:
    interval = 1.0
    end = '\r' if fd.isatty() else '\n'
    while True:
        sim_index, now, t_stop, timescale = env.get_progress()
        _print_progress(sim_index, now, t_stop, timescale, end=end, fd=fd)
        t0 = timeit.default_timer()
        yield env.timeout(interval)
        t1 = timeit.default_timer()
        interval *= period_s / (t1 - t0)


def _print_progress(
    sim_index: Optional[int],
    now: Union[int, float],
    t_stop: Optional[Union[int, float]],
    timescale: TimeValue,
    end: str,
    fd: IO,
) -> None:
    parts = []
    if sim_index:
        parts.append(f'Sim {sim_index}')
    magnitude, units = timescale
    if magnitude == 1:
        parts.append(f'{now:6.0f} {units}')
    else:
        parts.append(f'{magnitude}x{now:6.0f} {units}')
    if t_stop:
        parts.append(f'({100 * now / t_stop:.0f}%)')
    else:
        parts.append('(N/A%)')
    print(*parts, end=end, file=fd)
    fd.flush()


def _get_standalone_pbar(
    env: 'SimEnvironment', max_width: int, fd: IO
) -> progressbar.ProgressBar:
    pbar = progressbar.ProgressBar(
        fd=fd,
        min_value=0,
        max_value=progressbar.UnknownLength,
        widgets=_get_progressbar_widgets(
            env.sim_index, env.timescale, know_stop_time=False
        ),
    )

    if max_width and pbar.term_width > max_width:
        pbar.term_width = max_width

    return pbar


def _standalone_pbar_process(
    env: 'SimEnvironment', pbar: progressbar.ProgressBar, period_s: Union[int, float]
) -> Generator[simpy.Timeout, None, None]:
    interval = 1.0
    while True:
        sim_index, now, t_stop, timescale = env.get_progress()
        if t_stop and pbar.max_value != t_stop:
            pbar.max_value = t_stop
            pbar.widgets = _get_progressbar_widgets(
                sim_index, timescale, know_stop_time=True
            )
        pbar.update(now)
        t0 = timeit.default_timer()
        yield env.timeout(interval)
        t1 = timeit.default_timer()
        interval *= period_s / (t1 - t0)


def _get_progressbar_widgets(
    sim_index: Optional[int], timescale: TimeValue, know_stop_time: bool
) -> List[progressbar.widgets.WidgetBase]:
    widgets = []

    if sim_index is not None:
        widgets.append(f'Sim {sim_index:3}|')

    magnitude, units = timescale
    if magnitude == 1:
        sim_time_format = f'%(value)6.0f {units}|'
    else:
        sim_time_format = f'{magnitude}x%(value)6.0f {units}|'
    widgets.append(progressbar.FormatLabel(sim_time_format))

    widgets.append(progressbar.Percentage())

    if know_stop_time:
        widgets.append(progressbar.Bar())
    else:
        widgets.append(progressbar.BouncingBar())

    widgets.append(progressbar.ETA())

    return widgets


def get_multi_progress_manager(progress_queue: Optional['Queue[ProgressTuple]']):
    @contextmanager
    def progress_producer(env):
        if progress_queue:
            period_s = _get_interval_period_s(env.config)
            env.process(_progress_enqueue_process(env, period_s, progress_queue))
            try:
                yield None
            finally:
                progress_queue.put((env.sim_index, env.now, env.now, env.timescale))
        else:
            yield None

    return progress_producer


def _progress_enqueue_process(
    env: 'SimEnvironment',
    period_s: Union[int, float],
    progress_queue: 'Queue[ProgressTuple]',
) -> Generator[simpy.Timeout, None, None]:
    interval = 1.0
    while True:
        progress_queue.put(env.get_progress())
        t0 = timeit.default_timer()
        yield env.timeout(interval)
        t1 = timeit.default_timer()
        interval *= period_s / (t1 - t0)


def consume_multi_progress(
    progress_queue: 'Queue[ProgressTuple]',
    num_workers: int,
    num_simulations: int,
    max_width: int,
) -> None:
    fd = sys.stderr
    try:
        if fd.isatty():
            if progressbar and colorama:
                _consume_multi_display_multi_pbar(
                    progress_queue, num_workers, num_simulations, max_width, fd
                )
            elif progressbar:
                _consume_multi_display_single_pbar(
                    progress_queue, num_workers, num_simulations, max_width, fd
                )
            else:
                _consume_multi_display_simple(
                    progress_queue, num_workers, num_simulations, max_width, fd
                )
        else:
            _consume_multi_display_simple(
                progress_queue, num_workers, num_simulations, max_width, fd
            )
    except KeyboardInterrupt:
        pass


def _consume_multi_display_simple(
    progress_queue: 'Queue[ProgressTuple]',
    num_workers: int,
    num_simulations: int,
    max_width: int,
    fd: IO,
) -> None:
    start_date = datetime.now()
    isatty = fd.isatty()
    end = '\r' if isatty else '\n'
    try:
        completed: Set[Optional[int]] = set()
        _print_simple(len(completed), num_simulations, timedelta(), end, fd)
        last_print_date = start_date
        while len(completed) < num_simulations:
            progress: ProgressTuple = progress_queue.get()  # type: ignore
            sim_index, now, t_stop, timescale = progress
            now_date = datetime.now()
            td = now_date - start_date
            td_print = now_date - last_print_date
            if now == t_stop:
                completed.add(sim_index)
                _print_simple(len(completed), num_simulations, td, end, fd)
                last_print_date = now_date
            elif isatty and td_print.total_seconds() >= 1:
                _print_simple(len(completed), num_simulations, td, end, fd)
                last_print_date = now_date
    finally:
        if isatty:
            print(file=fd)


def _print_simple(
    num_completed: int, num_simulations: int, td: timedelta, end: str, fd: IO
) -> None:
    if fd.closed:
        return
    print(
        timedelta(td.days, td.seconds),
        num_completed,
        'of',
        num_simulations,
        'simulations',
        f'({num_completed / num_simulations:.0%})',
        end=end,
        file=fd,
    )
    fd.flush()


def _consume_multi_display_single_pbar(
    progress_queue: 'Queue[ProgressTuple]',
    num_workers: int,
    num_simulations: int,
    max_width: int,
    fd: IO,
):
    overall_pbar = _get_overall_pbar(num_simulations, max_width, fd=fd)
    try:
        completed: Set[Optional[int]] = set()
        while len(completed) < num_simulations:
            progress: ProgressTuple = progress_queue.get()  # type: ignore
            sim_index, now, t_stop, timescale = progress
            if now == t_stop:
                completed.add(sim_index)
                overall_pbar.update(len(completed))
    finally:
        overall_pbar.finish()


def _consume_multi_display_multi_pbar(
    progress_queue: 'Queue[ProgressTuple]',
    num_workers: int,
    num_simulations: int,
    max_width: int,
    fd: IO,
) -> None:
    # In order to display multiple progress bars, we need to manipulate the
    # terminal/console to move up lines. Colorama is used to wrap stderr such
    # that ANSI escape sequences are mapped to equivalent win32 API calls.
    fd = colorama.AnsiToWin32(fd).stream

    def ansi_up(n):
        return b'\x1b[{}A'.decode('latin1').format(n)

    ansi_bold = b'\x1b[1m'.decode('latin1')
    ansi_norm = b'\x1b[0m'.decode('latin1')

    overall_pbar = _get_overall_pbar(num_simulations, max_width, fd)

    try:
        worker_progress: Dict[Optional[int], progressbar.ProgressBar] = {}
        completed: Set[Optional[int]] = set()
        while len(completed) < num_simulations:
            progress: ProgressTuple = progress_queue.get()  # type: ignore
            sim_index, now, t_stop, timescale = progress

            if now == t_stop:
                completed.add(sim_index)

            if worker_progress:
                print(ansi_up(len(worker_progress)), end='', file=fd)

            if sim_index in worker_progress:
                for pindex, pbar in worker_progress.items():
                    if sim_index == pindex and pbar:
                        if now == t_stop:
                            pbar.finish()
                            worker_progress[sim_index] = None
                        else:
                            if t_stop and pbar.max_value != t_stop:
                                pbar.max_value = t_stop
                                pbar.widgets = _get_progressbar_widgets(
                                    sim_index, timescale, know_stop_time=True
                                )
                            pbar.update(now)
                            print(file=fd)
                    else:
                        print(file=fd)
            else:
                for pindex, pbar in worker_progress.items():
                    if pbar is None:
                        worker_progress.pop(pindex)
                        break
                print('\n' * len(worker_progress), file=fd)
                pbar = progressbar.ProgressBar(
                    fd=fd,
                    term_width=overall_pbar.term_width,
                    min_value=0,
                    max_value=(progressbar.UnknownLength if t_stop is None else t_stop),
                    widgets=_get_progressbar_widgets(
                        sim_index, timescale, know_stop_time=t_stop is not None
                    ),
                )
                worker_progress[sim_index] = pbar

            print(ansi_bold, end='', file=fd)
            overall_pbar.update(len(completed))
            print(ansi_norm, end='', file=fd)
    finally:
        print(ansi_bold, end='', file=fd)
        overall_pbar.finish()
        print(ansi_norm, end='', file=fd)


def _get_overall_pbar(
    num_simulations: int, max_width: int, fd: IO
) -> progressbar.ProgressBar:
    pbar = progressbar.ProgressBar(
        fd=fd,
        min_value=0,
        max_value=num_simulations,
        widgets=[
            progressbar.FormatLabel('%(value)s of %(max_value)s '),
            'simulations (',
            progressbar.Percentage(),
            ') ',
            progressbar.Bar(),
            progressbar.ETA(),
        ],
    )

    if max_width and pbar.term_width > max_width:
        pbar.term_width = max_width

    return pbar



"""Simulation model with batteries included."""

from contextlib import closing
from multiprocessing import Process, Queue, cpu_count
from pprint import pprint
from threading import Thread
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
)
import json
import os
import random
import shutil
import timeit

import simpy
import yaml

from desmod.config import ConfigDict, ConfigFactor, factorial_config
from desmod.progress import (
    ProgressTuple,
    consume_multi_progress,
    get_multi_progress_manager,
    standalone_progress_manager,
)
from desmod.timescale import parse_time, scale_time
from desmod.tracer import TraceManager

if TYPE_CHECKING:
    from desmod.component import Component  # noqa: F401

ResultDict = Dict[str, Any]


class SimEnvironment(simpy.Environment):
    """Simulation Environment.

    The :class:`SimEnvironment` class is a :class:`simpy.Environment` subclass
    that adds some useful features:

     - Access to the configuration dictionary (`config`).
     - Access to a seeded pseudo-random number generator (`rand`).
     - Access to the simulation timescale (`timescale`).
     - Access to the simulation duration (`duration`).

    Some models may need to share additional state with all its
    :class:`desmod.component.Component` instances. SimEnvironment may be
    subclassed to add additional members to achieve this sharing.

    :param dict config: A fully-initialized configuration dictionary.

    """

    def __init__(self, config: ConfigDict) -> None:
        super().__init__()
        #: The configuration dictionary.
        self.config = config

        #: The pseudo-random number generator; an instance of
        #: :class:`random.Random`.
        self.rand = random.Random()
        seed = config.setdefault('sim.seed', None)
        self.rand.seed(seed, version=1)

        timescale_str = self.config.setdefault('sim.timescale', '1 s')

        #: Simulation timescale ``(magnitude, units)`` tuple. The current
        #: simulation time is ``now * timescale``.
        self.timescale = parse_time(timescale_str)

        duration = config.setdefault('sim.duration', '0 s')

        #: The intended simulation duration, in units of :attr:`timescale`.
        self.duration = scale_time(parse_time(duration), self.timescale)

        #: The simulation runs "until" this event. By default, this is the
        #: configured "sim.duration", but may be overridden by subclasses.
        self.until = self.duration

        #: From 'meta.sim.index', the simulation's index when running multiple
        #: related simulations or `None` for a standalone simulation.
        self.sim_index: Optional[int] = config.get('meta.sim.index')

        #: :class:`TraceManager` instance.
        self.tracemgr = TraceManager(self)

    def time(self, t: Optional[float] = None, unit: str = 's') -> Union[int, float]:
        """The current simulation time scaled to specified unit.

        :param float t: Time in simulation units. Default is :attr:`now`.
        :param str unit: Unit of time to scale to. Default is 's' (seconds).
        :returns: Simulation time scaled to to `unit`.

        """
        target_scale = parse_time(unit)
        ts_mag, ts_unit = self.timescale
        sim_time = ((self.now if t is None else t) * ts_mag, ts_unit)
        return scale_time(sim_time, target_scale)

    def get_progress(self) -> ProgressTuple:
        if isinstance(self.until, SimStopEvent):
            t_stop = self.until.t_stop
        else:
            t_stop = self.until
        return self.sim_index, self.now, t_stop, self.timescale


class SimStopEvent(simpy.Event):
    """Event appropriate for stopping the simulation.

    An instance of this event may be used to override `SimEnvironment.until` to
    dynamically choose when to stop the simulation. The simulation may be
    stopped by calling :meth:`schedule()`. The optional `delay` parameter may
    be used to schedule the simulation to stop at an offset from the current
    simulation time.

    """

    def __init__(self, env: SimEnvironment) -> None:
        super().__init__(env)
        self.t_stop: Optional[Union[int, float]] = None

    def schedule(self, delay: Union[int, float] = 0) -> None:
        assert not self.triggered
        assert delay >= 0
        self._ok = True
        self._value = None
        self.env.schedule(self, simpy.events.URGENT, delay)
        self.t_stop = self.env.now + delay


class _Workspace:
    """Context manager for workspace directory management."""

    def __init__(self, config: ConfigDict) -> None:
        self.workspace: str = config.setdefault(
            'meta.sim.workspace', config.setdefault('sim.workspace', os.curdir)
        )
        self.overwrite: bool = config.setdefault('sim.workspace.overwrite', False)
        self.prev_dir: str = os.getcwd()

    def __enter__(self) -> '_Workspace':
        if os.path.relpath(self.workspace) != os.curdir:
            workspace_exists = os.path.isdir(self.workspace)
            if self.overwrite and workspace_exists:
                shutil.rmtree(self.workspace)
            if self.overwrite or not workspace_exists:
                os.makedirs(self.workspace)
            os.chdir(self.workspace)
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        os.chdir(self.prev_dir)
        return None


def simulate(
    config: ConfigDict,
    top_type: Type['Component'],
    env_type: Type[SimEnvironment] = SimEnvironment,
    reraise: bool = True,
    progress_manager=standalone_progress_manager,
) -> ResultDict:
    """Initialize, elaborate, and run a simulation.

     All exceptions are caught by `simulate()` so they can be logged and
     captured in the result file. By default, any unhandled exception caught by
     `simulate()` will be re-raised. Setting `reraise` to False prevents
     exceptions from propagating to the caller. Instead, the returned result
     dict will indicate if an exception occurred via the 'sim.exception' item.

    :param dict config: Configuration dictionary for the simulation.
    :param top_type: The model's top-level Component subclass.
    :param env_type: :class:`SimEnvironment` subclass.
    :param bool reraise: Should unhandled exceptions propogate to the caller.
    :returns:
        Dictionary containing the model-specific results of the simulation.
    """
    t0 = timeit.default_timer()
    result: ResultDict = {}
    result_file = config.setdefault('sim.result.file')
    config_file = config.setdefault('sim.config.file')
    try:
        with _Workspace(config):
            env = env_type(config)
            with closing(env.tracemgr):
                try:
                    top_type.pre_init(env)
                    env.tracemgr.flush()
                    with progress_manager(env):
                        top = top_type(parent=None, env=env)
                        top.elaborate()
                        env.tracemgr.flush()
                        env.run(until=env.until)
                        env.tracemgr.flush()
                        top.post_simulate()
                        env.tracemgr.flush()
                        top.get_result(result)
                except BaseException as e:
                    env.tracemgr.trace_exception()
                    result['sim.exception'] = repr(e)
                    raise
                else:
                    result['sim.exception'] = None
                finally:
                    env.tracemgr.flush()
                    result['config'] = config
                    result['sim.now'] = env.now
                    result['sim.time'] = env.time()
                    result['sim.runtime'] = timeit.default_timer() - t0
                    _dump_dict(config_file, config)
                    _dump_dict(result_file, result)
    except BaseException as e:
        if reraise:
            raise
        result.setdefault('config', config)
        result.setdefault('sim.runtime', timeit.default_timer() - t0)
        if result.get('sim.exception') is None:
            result['sim.exception'] = repr(e)
    return result


def simulate_factors(
    base_config: ConfigDict,
    factors: List[ConfigFactor],
    top_type: Type['Component'],
    env_type: Type[SimEnvironment] = SimEnvironment,
    jobs: Optional[int] = None,
    config_filter: Optional[Callable[[ConfigDict], bool]] = None,
) -> List[ResultDict]:
    """Run multi-factor simulations in separate processes.

    The `factors` are used to compose specialized config dictionaries for the
    simulations.

    The :mod:`python:multiprocessing` module is used run each simulation with a
    separate Python process. This allows multi-factor simulations to run in
    parallel on all available CPU cores.

    :param dict base_config: Base configuration dictionary to be specialized.
    :param list factors: List of factors.
    :param top_type: The model's top-level Component subclass.
    :param env_type: :class:`SimEnvironment` subclass.
    :param int jobs: User specified number of concurent processes.
    :param function config_filter:
        A function which will be passed a config and returns a bool to filter.
    :returns: Sequence of result dictionaries for each simulation.

    """
    configs = list(factorial_config(base_config, factors, 'meta.sim.special'))
    ws = base_config.setdefault('sim.workspace', os.curdir)
    overwrite = base_config.setdefault('sim.workspace.overwrite', False)

    for index, config in enumerate(configs):
        config['meta.sim.index'] = index
        config['meta.sim.workspace'] = os.path.join(ws, str(index))
    if config_filter is not None:
        configs[:] = filter(config_filter, configs)
    if overwrite and os.path.relpath(ws) != os.curdir and os.path.isdir(ws):
        shutil.rmtree(ws)
    return simulate_many(configs, top_type, env_type, jobs)


def simulate_many(
    configs: Sequence[ConfigDict],
    top_type: Type['Component'],
    env_type: Type[SimEnvironment] = SimEnvironment,
    jobs: Optional[int] = None,
) -> List[ResultDict]:
    """Run multiple experiments in separate processes.

    The :mod:`python:multiprocessing` module is used run each simulation with a
    separate Python process. This allows multi-factor simulations to run in
    parallel on all available CPU cores.

    :param dict configs: list of configuration dictionary for the simulation.
    :param top_type: The model's top-level Component subclass.
    :param env_type: :class:`SimEnvironment` subclass.
    :param int jobs: User specified number of concurent processes.
    :returns: Sequence of result dictionaries for each simulation.

    """
    if jobs is not None and jobs < 1:
        raise ValueError(f'Invalid number of jobs: {jobs}')

    progress_enable = any(
        config.setdefault('sim.progress.enable', False) for config in configs
    )

    progress_queue: Optional[
        Queue[ProgressTuple]
    ] = Queue() if progress_enable else None
    result_queue: Queue[ResultDict] = Queue()
    config_queue: Queue[Optional[ConfigDict]] = Queue()

    workspaces = set()
    max_width = 0
    for index, config in enumerate(configs):
        max_width = max(config.setdefault('sim.progress.max_width', 0), max_width)

        workspace = os.path.normpath(
            config.setdefault(
                'meta.sim.workspace', config.setdefault('sim.workspace', os.curdir)
            )
        )
        if workspace in workspaces:
            raise ValueError(f'Duplicate workspace: {workspace}')
        workspaces.add(workspace)

        config.setdefault('meta.sim.index', index)
        config['sim.progress.enable'] = progress_enable
        config_queue.put(config)

    num_workers = min(len(configs), cpu_count())
    if jobs is not None:
        num_workers = min(num_workers, jobs)

    workers = []
    for i in range(num_workers):
        worker = Process(
            name=f'sim-worker-{i}',
            target=_simulate_worker,
            args=(
                top_type,
                env_type,
                False,
                progress_queue,
                config_queue,
                result_queue,
            ),
        )
        worker.daemon = True  # Workers die if main process dies.
        worker.start()
        workers.append(worker)
        config_queue.put(None)  # A stop sentinel for each worker.

    if progress_enable:
        progress_thread = Thread(
            target=consume_multi_progress,
            args=(progress_queue, num_workers, len(configs), max_width),
        )
        progress_thread.daemon = True
        progress_thread.start()

    results = [result_queue.get() for _ in configs]

    if progress_enable:
        # Although this is a daemon thread, we still make a token attempt to
        # join with it. This avoids a race with certain testing frameworks
        # (ahem, py.test) that may monkey-patch and close stderr while
        # progress_thread is still using it.
        progress_thread.join(1)

    for worker in workers:
        worker.join(5)

    return sorted(results, key=lambda r: r['config']['meta.sim.index'])


def _simulate_worker(
    top_type: Type['Component'],
    env_type: Type[SimEnvironment],
    reraise: bool,
    progress_queue: Optional['Queue[ProgressTuple]'],
    config_queue: 'Queue[Optional[ConfigDict]]',
    result_queue: 'Queue[ResultDict]',
):
    progress_manager = get_multi_progress_manager(progress_queue)
    while True:
        config = config_queue.get()
        if config is None:
            break
        result = simulate(config, top_type, env_type, reraise, progress_manager)
        result_queue.put(result)


def _dump_dict(filename: str, dump_dict: Dict[str, Any]):
    if filename is not None:
        _, ext = os.path.splitext(filename)
        if ext not in ['.yaml', '.yml', '.json', '.py']:
            raise ValueError(f'Invalid extension: {ext}')
        with open(filename, 'w') as dump_file:
            if ext in ['.yaml', '.yml']:
                yaml.safe_dump(dump_dict, stream=dump_file)
            elif ext == '.json':
                json.dump(dump_dict, dump_file, sort_keys=True, indent=2)
            else:
                assert ext == '.py'
                pprint(dump_dict, stream=dump_file)



import pytest

from desmod.util import partial_format


@pytest.mark.parametrize(
    'expected, format_str, kwargs',
    [
        ('abc', 'abc', {}),
        ('aBc', 'a{b}c', {'b': 'B'}),
        ('a{b!r}c', 'a{b!r}c', {}),
        ("a'B'c", 'a{b!r}c', {'b': 'B'}),
        ('a {:.2f} c', 'a {:.{digits}f} c', {'digits': 2}),
        ('A{b}C', '{a}{b}{c}', {'a': 'A', 'c': 'C'}),
    ],
)
def test_partial_format(expected, format_str, kwargs):
    assert expected == partial_format(format_str, **kwargs)
